"""
Flitter language compiler unit tests
"""

import unittest

from flitter.model import Vector, Node, null
from flitter.language import functions
from flitter.language.tree import (Literal, Name, Sequence,
                                   Positive, Negative, Ceil, Floor, Fract, Power,
                                   Add, Subtract, Multiply, Divide, FloorDivide, Modulo,
                                   EqualTo, NotEqualTo, LessThan, GreaterThan, LessThanOrEqualTo, GreaterThanOrEqualTo,
                                   Not, And, Or, Xor, Range, Slice, Lookup,
                                   Tag, Attributes, Append,
                                   Let, Call, For, IfElse,
                                   Import, Function, Export, Top,
                                   Binding, PolyBinding, IfCondition)
from flitter.language.vm import Program


class CompilerTestCase(unittest.TestCase):
    def assertCompilesTo(self, expr, program, lnames=(), with_errors=None):
        program.optimize()
        program.link()
        compiled = expr.compile(initial_lnames=lnames, log_errors=False)
        self.assertEqual(str(compiled), str(program))
        self.assertEqual(compiled.compiler_errors, set() if with_errors is None else with_errors)


class TestLiteral(CompilerTestCase):
    def test_null(self):
        self.assertCompilesTo(Literal(null), Program().literal(null))

    def test_vector(self):
        self.assertCompilesTo(Literal([1, 2, 3]), Program().literal([1, 2, 3]))

    def test_builtin(self):
        self.assertCompilesTo(Literal(functions.sqrtv), Program().literal(functions.sqrtv))


class TestName(CompilerTestCase):
    def test_undefined(self):
        self.assertCompilesTo(Name('z'), Program().literal(null), with_errors={"Unbound name 'z'"})

    def test_lname(self):
        self.assertCompilesTo(Name('x'), Program().local_load(0), lnames=('x',))
        self.assertCompilesTo(Name('x'), Program().local_load(1), lnames=('x', 'y'))

    def test_static_builtin(self):
        self.assertCompilesTo(Name('null'), Program().literal(null))
        self.assertCompilesTo(Name('sqrt'), Program().literal(functions.sqrtv))

    def test_dynamic_builtin(self):
        self.assertCompilesTo(Name('debug'), Program().literal(functions.debug))


class TestSequence(CompilerTestCase):
    def test_empty(self):
        self.assertCompilesTo(Sequence(()), Program().literal(null))

    def test_single(self):
        self.assertCompilesTo(Sequence((Name('x'),)), Program().local_load(0), lnames=('x',))

    def test_multiple(self):
        self.assertCompilesTo(Sequence((Name('x'), Name('y'))), Program().local_load(1).local_load(0).compose(2), lnames=('x', 'y'))


class TestUnaryExpressions(CompilerTestCase):
    def test_positive(self):
        self.assertCompilesTo(Positive(Name('x')), Program().local_load(0).pos(), lnames=('x',))

    def test_negative(self):
        self.assertCompilesTo(Negative(Name('x')), Program().local_load(0).neg(), lnames=('x',))

    def test_ceiling(self):
        self.assertCompilesTo(Ceil(Name('x')), Program().local_load(0).ceil(), lnames=('x',))

    def test_floor(self):
        self.assertCompilesTo(Floor(Name('x')), Program().local_load(0).floor(), lnames=('x',))

    def test_fract(self):
        self.assertCompilesTo(Fract(Name('x')), Program().local_load(0).fract(), lnames=('x',))

    def test_not(self):
        self.assertCompilesTo(Not(Name('x')), Program().local_load(0).not_(), lnames=('x',))


class TestBinaryExpressions(CompilerTestCase):
    def test_add(self):
        self.assertCompilesTo(Add(Name('x'), Name('y')), Program().local_load(1).local_load(0).add(), lnames=('x', 'y'))

    def test_subtract(self):
        self.assertCompilesTo(Subtract(Name('x'), Name('y')), Program().local_load(1).local_load(0).sub(), lnames=('x', 'y'))

    def test_multiply(self):
        self.assertCompilesTo(Multiply(Name('x'), Name('y')), Program().local_load(1).local_load(0).mul(), lnames=('x', 'y'))

    def test_divide(self):
        self.assertCompilesTo(Divide(Name('x'), Name('y')), Program().local_load(1).local_load(0).truediv(), lnames=('x', 'y'))

    def test_floor_divide(self):
        self.assertCompilesTo(FloorDivide(Name('x'), Name('y')), Program().local_load(1).local_load(0).floordiv(), lnames=('x', 'y'))

    def test_modulo(self):
        self.assertCompilesTo(Modulo(Name('x'), Name('y')), Program().local_load(1).local_load(0).mod(), lnames=('x', 'y'))

    def test_power(self):
        self.assertCompilesTo(Power(Name('x'), Name('y')), Program().local_load(1).local_load(0).pow(), lnames=('x', 'y'))

    def test_equal_to(self):
        self.assertCompilesTo(EqualTo(Name('x'), Name('y')), Program().local_load(1).local_load(0).eq(), lnames=('x', 'y'))

    def test_not_equal_to(self):
        self.assertCompilesTo(NotEqualTo(Name('x'), Name('y')), Program().local_load(1).local_load(0).ne(), lnames=('x', 'y'))

    def test_less_than(self):
        self.assertCompilesTo(LessThan(Name('x'), Name('y')), Program().local_load(1).local_load(0).lt(), lnames=('x', 'y'))

    def test_greater_than(self):
        self.assertCompilesTo(GreaterThan(Name('x'), Name('y')), Program().local_load(1).local_load(0).gt(), lnames=('x', 'y'))

    def test_less_than_or_equal_to(self):
        self.assertCompilesTo(LessThanOrEqualTo(Name('x'), Name('y')), Program().local_load(1).local_load(0).le(), lnames=('x', 'y'))

    def test_greater_than_or_equal_to(self):
        self.assertCompilesTo(GreaterThanOrEqualTo(Name('x'), Name('y')), Program().local_load(1).local_load(0).ge(), lnames=('x', 'y'))

    def test_xor(self):
        self.assertCompilesTo(Xor(Name('x'), Name('y')), Program().local_load(1).local_load(0).xor(), lnames=('x', 'y'))


class TestShortcutLogic(CompilerTestCase):
    def test_and(self):
        program = Program()
        END = program.new_label()
        program.local_load(1)
        program.dup()
        program.branch_false(END)
        program.drop()
        program.local_load(0)
        program.label(END)
        self.assertCompilesTo(And(Name('x'), Name('y')), program, lnames=('x', 'y'))

    def test_or(self):
        program = Program()
        END = program.new_label()
        program.local_load(1)
        program.dup()
        program.branch_true(END)
        program.drop()
        program.local_load(0)
        program.label(END)
        self.assertCompilesTo(Or(Name('x'), Name('y')), program, lnames=('x', 'y'))


class TestTernaryExpressions(CompilerTestCase):
    def test_range(self):
        self.assertCompilesTo(Range(Name('x'), Name('y'), Name('z')), Program().local_load(2).local_load(1).local_load(0).range(), lnames=('x', 'y', 'z'))


class TestSlice(CompilerTestCase):
    def test_dynamic(self):
        self.assertCompilesTo(Slice(Name('x'), Name('y')), Program().local_load(1).local_load(0).slice(), lnames=('x', 'y'))

    def test_literal(self):
        self.assertCompilesTo(Slice(Name('x'), Literal(0)), Program().local_load(0).slice_literal(Vector(0)), lnames=('x',))


class TestLookup(CompilerTestCase):
    def test_dynamic(self):
        self.assertCompilesTo(Lookup(Name('x')), Program().local_load(0).lookup(), lnames=('x',))

    def test_literal(self):
        self.assertCompilesTo(Lookup(Literal(Vector.symbol('x'))), Program().lookup_literal(Vector.symbol('x')))


class TestNodeOperations(CompilerTestCase):
    def test_tag(self):
        self.assertCompilesTo(Tag(Name('x'), 'y'), Program().local_load(0).tag('y'), lnames=('x',))

    def test_single_attribute(self):
        self.assertCompilesTo(Attributes(Name('x'), (Binding('y', Name('y')),)), Program().local_load(1).local_load(0).attribute('y'), lnames=('x', 'y'))

    def test_multiple_attributes(self):
        self.assertCompilesTo(Attributes(Name('x'), (Binding('y', Name('y')), Binding('z', Name('z')))),
                              Program().local_load(2).local_load(1).attribute('y').local_load(0).attribute('z'), lnames=('x', 'y', 'z'))

    def test_append(self):
        self.assertCompilesTo(Append(Name('x'), Name('y')), Program().local_load(1).local_load(0).append(), lnames=('x', 'y'))


class TestLet(CompilerTestCase):
    def test_single(self):
        self.assertCompilesTo(Let((PolyBinding(('x',), Literal(5)),), Name('x')),
                              Program().literal(5).local_push(1).local_load(0).local_drop(1))

    def test_multiple(self):
        self.assertCompilesTo(Let((PolyBinding(('x',), Literal(5)), PolyBinding(('y',), Literal(10))), Add(Name('x'), Name('y'))),
                              Program().literal(5).local_push(1).literal(10).local_push(1).local_load(1).local_load(0).add().local_drop(2))

    def test_multi_binding(self):
        self.assertCompilesTo(Let((PolyBinding(('x', 'y'), Literal([5, 10])),), Add(Name('x'), Name('y'))),
                              Program().literal([5, 10]).local_push(2).local_load(1).local_load(0).add().local_drop(2))


class TestCall(CompilerTestCase):
    def test_builtin_single_arg(self):
        self.assertCompilesTo(Call(Literal(functions.sqrtv), (Literal(25),)),
                              Program().literal(25).call_fast(functions.sqrtv, 1))

    def test_builtin_multiple_args(self):
        self.assertCompilesTo(Call(Literal(functions.clamp), (Name('x'), Literal(0), Literal(1))),
                              Program().local_load(0).literal(0).literal(1).call_fast(functions.clamp, 3), lnames=('x',))

    def test_builtin_keyword_arg_only(self):
        self.assertCompilesTo(Call(Literal(functions.sqrtv), (), (Binding('xs', Literal(25)),)),
                              Program().literal(25).literal(functions.sqrtv).call(0, ('xs',)))

    def test_builtin_multiple_keyword_args(self):
        self.assertCompilesTo(Call(Literal(functions.clamp), (), (Binding('xs', Name('x')), Binding('ys', Literal(0)), Binding('zs', Literal(1)))),
                              Program().local_load(0).literal(0).literal(1).literal(functions.clamp).call(0, ('xs', 'ys', 'zs')), lnames=('x',))

    def test_builtin_mixed_positional_and_keyword_args(self):
        self.assertCompilesTo(Call(Literal(functions.clamp), (Name('x'),), (Binding('ys', Literal(0)), Binding('zs', Literal(1)))),
                              Program().local_load(0).literal(0).literal(1).literal(functions.clamp).call(1, ('ys', 'zs')), lnames=('x',))

    def test_func(self):
        func = Function('func', (Binding('x', Literal(null)),), Name('x')).simplify()
        self.assertCompilesTo(Call(Literal(func), (Name('x'),)),
                              Program().local_load(0).literal(func).call(1), lnames=('x',))

    def test_multiple_builtins(self):
        self.assertCompilesTo(Call(Literal([functions.cosv, functions.sinv]), (Name('th'),)),
                              Program().local_load(0).literal([functions.cosv, functions.sinv]).call(1), lnames=('th',))


class TestFor(CompilerTestCase):
    def test_single_binding(self):
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.literal([1, 2, 3])
        program.begin_for(1)
        program.label(START)
        program.next(END)
        program.local_load(0)
        program.literal(2)
        program.mul()
        program.jump(START)
        program.label(END)
        program.end_for()
        self.assertCompilesTo(For(('x',), Literal([1, 2, 3]), Multiply(Name('x'), Literal(2))), program)

    def test_multi_binding(self):
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.literal([1, 2, 3, 4])
        program.begin_for(2)
        program.label(START)
        program.next(END)
        program.local_load(1)
        program.local_load(0)
        program.mul()
        program.jump(START)
        program.label(END)
        program.end_for()
        self.assertCompilesTo(For(('x', 'y'), Literal([1, 2, 3, 4]), Multiply(Name('x'), Name('y'))), program)


class TestIfElse(CompilerTestCase):
    def test_single_if_no_else(self):
        program = Program()
        END = program.new_label()
        program.local_load(1)
        NEXT = program.new_label()
        program.branch_false(NEXT)
        program.local_load(0)
        program.jump(END)
        program.label(NEXT)
        program.literal(null)
        program.label(END)
        self.assertCompilesTo(IfElse((IfCondition(Name('x'), Name('y')),), None), program, lnames=('x', 'y'))

    def test_single_if_else(self):
        program = Program()
        END = program.new_label()
        program.local_load(2)
        NEXT = program.new_label()
        program.branch_false(NEXT)
        program.local_load(1)
        program.jump(END)
        program.label(NEXT)
        program.local_load(0)
        program.label(END)
        self.assertCompilesTo(IfElse((IfCondition(Name('x'), Name('y')),), Name('z')), program, lnames=('x', 'y', 'z'))

    def test_multiple_if_else(self):
        program = Program()
        END = program.new_label()
        program.local_load(2)
        NEXT = program.new_label()
        program.branch_false(NEXT)
        program.local_load(1)
        program.jump(END)
        program.label(NEXT)
        program.local_load(4)
        NEXT = program.new_label()
        program.branch_false(NEXT)
        program.local_load(3)
        program.jump(END)
        program.label(NEXT)
        program.local_load(0)
        program.label(END)
        self.assertCompilesTo(IfElse((IfCondition(Name('x'), Name('y')), IfCondition(Name('a'), Name('b'))), Name('z')), program,
                              lnames=('a', 'b', 'x', 'y', 'z'))


class TestImport(CompilerTestCase):
    def test_import(self):
        self.assertCompilesTo(Import(('x', 'y'), Literal('module.fl'), Literal(null)),
                              Program().literal('module.fl').import_(('x', 'y')).literal(null).local_drop(2))


class TestFunction(CompilerTestCase):
    def test_single_parameter_no_captures(self):
        func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)), captures=())
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.literal(null)
        program.jump(END)
        program.label(START)
        program.local_load(0)
        program.literal(5)
        program.add()
        program.exit()
        program.label(END)
        program.func(START, 'func', ('x',), 0)
        self.assertCompilesTo(func, program)

    def test_unknown_captures(self):
        """If `captures` is `None` then *all* lnames are captured in case"""
        func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)))
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.local_load(1)
        program.local_load(0)
        program.literal(null)
        program.jump(END)
        program.label(START)
        program.local_load(0)
        program.literal(5)
        program.add()
        program.exit()
        program.label(END)
        program.func(START, 'func', ('x',), 2)
        self.assertCompilesTo(func, program, lnames=('a', 'b'))

    def test_two_parameters_no_captures(self):
        func = Function('func', (Binding('x', Literal(null)), Binding('y', Literal(5))), Add(Name('x'), Name('y')), captures=())
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.literal(null)
        program.literal(5)
        program.jump(END)
        program.label(START)
        program.local_load(1)
        program.local_load(0)
        program.add()
        program.exit()
        program.label(END)
        program.func(START, 'func', ('x', 'y'), 0)
        self.assertCompilesTo(func, program)

    def test_two_parameters_one_capture(self):
        func = Function('func', (Binding('x', Literal(null)), Binding('y', Literal(5))), Multiply(Add(Name('x'), Name('y')), Name('z')), captures=('z',))
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.local_load(0)
        program.literal(null)
        program.literal(5)
        program.jump(END)
        program.label(START)
        program.local_load(1)
        program.local_load(0)
        program.add()
        program.local_load(3)
        program.mul()
        program.exit()
        program.label(END)
        program.func(START, 'func', ('x', 'y'), 1)
        self.assertCompilesTo(func, program, lnames=('z',))

    def test_recursive_no_captures(self):
        func = Function('func',
                        (Binding('x', Literal(null)),),
                        IfElse((IfCondition(GreaterThan(Name('x'), Literal(0)), Add(Name('x'), Call(Name('func'), (Subtract(Name('x'), Literal(1)),)))),),
                               Literal(0)),
                        captures=())
        program = Program()
        START = program.new_label()
        END = program.new_label()
        program.literal(null)
        program.jump(END)
        program.label(START)
        LAST = program.new_label()
        program.local_load(0)
        program.literal(0)
        program.gt()
        NEXT = program.new_label()
        program.branch_false(NEXT)
        program.local_load(0)
        program.local_load(0)
        program.literal(1)
        program.sub()
        program.local_load(1)
        program.call(1)
        program.add()
        program.jump(LAST)
        program.label(NEXT)
        program.literal(0)
        program.label(LAST)
        program.exit()
        program.label(END)
        program.func(START, 'func', ('x',), 0)
        self.assertCompilesTo(func, program)


class TestExport(CompilerTestCase):
    def test_empty(self):
        self.assertCompilesTo(Export(), Program().literal(null))

    def test_explicit(self):
        self.assertCompilesTo(Export({'x': Vector(5)}), Program().literal(5).export('x').literal(null))

    def test_initial_lnames_ignored(self):
        self.assertCompilesTo(Export(), Program().literal(null), lnames=('x', 'y'))

    def test_single(self):
        self.assertCompilesTo(Let((PolyBinding(('x',), Literal(5)),), Export()),
                              Program().literal(5).local_push(1).local_load(0).export('x').literal(null).local_drop(1))

    def test_multiple(self):
        self.assertCompilesTo(Let((PolyBinding(('x',), Literal(5)), PolyBinding(('y',), Literal(10))), Export({'z': Vector(15)})),
                              Program().literal(5).local_push(1)
                                       .literal(10).local_push(1)
                                       .literal(15).export('z')
                                       .local_load(0).export('y')
                                       .local_load(1).export('x')
                                       .literal(null)
                                       .local_drop(2))


class TestTop(CompilerTestCase):
    def test_empty(self):
        self.assertCompilesTo(Top((), Sequence(())), Program().literal(null).append())

    def test_literal_node(self):
        self.assertCompilesTo(Top((), Literal(Node('window'))), Program().literal(Node('window')).append())



"""
Flitter window management
"""

import array
import ctypes
from collections import namedtuple
import importlib
import sys

import glfw
from loguru import logger
from mako.template import Template
import moderngl
import numpy as np

from ...clock import system_clock
from .glconstants import GL_RGBA8, GL_RGBA16F, GL_RGBA32F, GL_FRAMEBUFFER_SRGB
from .glsl import TemplateLoader
from ...model import Vector


def set_uniform_vector(uniform, vector):
    dtype = {'f': float, 'd': float, 'i': int, 'I': int}[uniform.fmt[-1]]
    n, m = uniform.array_length, uniform.dimension
    if (value := vector.match(n * m, dtype)) is not None:
        if m == 1:
            uniform.value = value if n == 1 else list(value)
        elif n == 1:
            uniform.value = tuple(value)
        else:
            uniform.value = [tuple(value[i*m:(i+1)*m]) for i in range(n)]


ColorFormat = namedtuple('ColorFormat', ('moderngl_dtype', 'gl_format'))

COLOR_FORMATS = {
    8: ColorFormat('f1', GL_RGBA8),
    16: ColorFormat('f2', GL_RGBA16F),
    32: ColorFormat('f4', GL_RGBA32F)
}

DEFAULT_LINEAR = False
DEFAULT_COLORBITS = 16
PUSHED = Vector.symbol('pushed')
RELEASED = Vector.symbol('released')
PUSHED_BEAT = Vector.symbol('pushed').concat(Vector.symbol('beat'))
RELEASED_BEAT = Vector.symbol('released').concat(Vector.symbol('beat'))

GLFUNCTYPE = ctypes.WINFUNCTYPE if sys.platform == 'win32' else ctypes.CFUNCTYPE


def get_scene_node_class(kind):
    global ClassCache
    if kind in ClassCache:
        return ClassCache[kind]
    try:
        module = importlib.import_module(f'.{kind}', __package__)
        cls = module.SCENE_NODE_CLASS
    except ModuleNotFoundError:
        logger.warning("No sub-module for '{}'", kind)
        cls = None
    except ImportError:
        logger.exception("Import error")
        cls = None
    except AttributeError:
        logger.warning("Sub-module '{}' does not contain a scene node class", kind)
        cls = None
    ClassCache[kind] = cls
    return cls


class SceneNode:
    def __init__(self, glctx):
        self.glctx = glctx
        self.children = []
        self.width = None
        self.height = None
        self.tags = set()
        self.hidden = False
        self._texture_data = None

    @property
    def name(self):
        return '#'.join((self.__class__.__name__.lower(), *self.tags))

    @property
    def texture(self):
        raise NotImplementedError()

    @property
    def texture_data(self):
        if self._texture_data is None and (texture := self.texture) is not None:
            dtype = {'f1': 'uint8', 'f2': 'float16', 'f4': 'float32'}[texture.dtype]
            data = np.ndarray((texture.height, texture.width, texture.components), dtype, texture.read())
            self._texture_data = data.astype('float64')
            if dtype == 'uint8':
                self._texture_data /= 255
        return self._texture_data

    @property
    def child_textures(self):
        textures = {}
        i = 0
        for child in self.children:
            if child.texture is not None and not child.hidden:
                textures[f'texture{i}'] = child.texture
                i += 1
        return textures

    def destroy(self):
        self.purge()
        self.release()
        self.glctx = None

    def purge(self):
        while self.children:
            self.children.pop().destroy()

    async def update(self, engine, node, default_size=(512, 512), **kwargs):
        self._texture_data = None
        references = kwargs.setdefault('references', {})
        if node_id := node.get('id', 1, str):
            references[node_id] = self
        self.hidden = node.get('hidden', 1, bool, False)
        resized = False
        width, height = node.get('size', 2, int, default_size)
        if width != self.width or height != self.height:
            self.width = width
            self.height = height
            resized = True
        self.tags = node.tags
        self.create(engine, node, resized, **kwargs)
        await self.descend(engine, node, **kwargs)
        self.render(node, **kwargs)

    def similar_to(self, node):
        return node.tags == self.tags

    async def descend(self, engine, node, **kwargs):
        existing = self.children
        updated = []
        for child in node.children:
            if self.handle_node(engine, child, **kwargs):
                continue
            cls = get_scene_node_class(child.kind)
            if cls is not None:
                for i, scene_node in enumerate(existing):
                    if type(scene_node) is cls and scene_node.similar_to(child):
                        scene_node = existing.pop(i)
                        break
                else:
                    scene_node = cls(self.glctx)
                await scene_node.update(engine, child, default_size=(self.width, self.height), **kwargs)
                updated.append(scene_node)
        while existing:
            existing.pop().destroy()
        self.children = updated

    def handle_node(self, engine, node, **kwargs):
        return False

    def create(self, engine, node, resized, **kwargs):
        pass

    def render(self, node, **kwargs):
        raise NotImplementedError()

    def release(self):
        raise NotImplementedError()


class Reference(SceneNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._reference = None

    @property
    def texture(self):
        return self._reference.texture if self._reference is not None else None

    async def update(self, engine, node, references=None, **kwargs):
        node_id = node.get('id', 1, str)
        self._reference = references.get(node_id) if references is not None and node_id else None

    def destroy(self):
        self._reference = None


class ProgramNode(SceneNode):
    DEFAULT_VERTEX_SOURCE = TemplateLoader.get_template('default.vert')
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('default.frag')
    CLEAR_COLOR = (0, 0, 0, 0)

    def __init__(self, glctx):
        super().__init__(glctx)
        self._program = None
        self._rectangle = None
        self._vertex_source = None
        self._fragment_source = None
        self._last = None

    def release(self):
        self._program = None
        self._rectangle = None
        self._vertex_source = None
        self._fragment_source = None
        self._last = None

    @property
    def framebuffer(self):
        raise NotImplementedError()

    @property
    def size(self):
        return self.framebuffer.size

    def create(self, engine, node, resized, **kwargs):
        super().create(engine, node, resized, **kwargs)
        if resized and self._last is not None:
            self._last = None

    def get_vertex_source(self, node):
        if 'vertex' in node:
            vertex = Template(node.get('vertex', 1, str))
        else:
            vertex = self.DEFAULT_VERTEX_SOURCE
        return vertex.render(HEADER=self.glctx.extra['HEADER'], child_textures=list(self.child_textures))

    def get_fragment_source(self, node):
        if 'fragment' in node:
            fragment = Template(node.get('fragment', 1, str))
        else:
            fragment = self.DEFAULT_FRAGMENT_SOURCE
        composite = node.get('composite', 1, str, node.get('blend', 1, str, 'over'))
        return fragment.render(HEADER=self.glctx.extra['HEADER'], child_textures=list(self.child_textures), composite=composite)

    def make_last(self):
        raise NotImplementedError()

    def compile(self, node):
        vertex_source = self.get_vertex_source(node)
        fragment_source = self.get_fragment_source(node)
        if vertex_source != self._vertex_source or fragment_source != self._fragment_source:
            self._vertex_source = vertex_source
            self._fragment_source = fragment_source
            self._program = None
            self._rectangle = None
            try:
                start = system_clock()
                self._program = self.glctx.program(vertex_shader=self._vertex_source, fragment_shader=self._fragment_source)
                vertices = self.glctx.buffer(array.array('f', [-1, 1, -1, -1, 1, 1, 1, -1]))
                self._rectangle = self.glctx.vertex_array(self._program, [(vertices, '2f', 'position')], mode=moderngl.TRIANGLE_STRIP)
            except Exception as exc:
                error = str(exc).strip().split('\n')
                logger.error("{} GL program compile failed: {}", self.name, error[-1])
                source = self._fragment_source if 'fragment_shader' in error else self._vertex_source
                source = '\n'.join(f'{i+1:3d}|{line}' for i, line in enumerate(source.split('\n')))
                logger.trace("Failing source:\n{}", source)
            else:
                end = system_clock()
                logger.debug("{} GL program compiled in {:.1f}ms", self.name, 1000 * (end - start))

    def render(self, node, **kwargs):
        self.compile(node)
        if self._rectangle is not None:
            passes = max(1, node.get('passes', 1, int, 1))
            sampler_args = {'repeat_x': False, 'repeat_y': False}
            if (border := node.get('border', 4, float)) is not None:
                sampler_args['border_color'] = tuple(border)
            elif (repeat := node.get('repeat', 2, bool)) is not None:
                if repeat[0]:
                    del sampler_args['repeat_x']
                if repeat[1]:
                    del sampler_args['repeat_y']
            child_textures = self.child_textures
            samplers = []
            unit = 1
            pass_member = None
            for name in self._program:
                member = self._program[name]
                if isinstance(member, moderngl.Uniform):
                    if name == 'last':
                        if self._last is None:
                            self._last = self.make_last()
                        if sampler_args:
                            sampler = self.glctx.sampler(texture=self._last, **sampler_args)
                            sampler.use(location=unit)
                            samplers.append(sampler)
                        else:
                            self._last.use(location=unit)
                        member.value = unit
                        unit += 1
                    elif name in child_textures:
                        if sampler_args:
                            sampler = self.glctx.sampler(texture=child_textures[name], **sampler_args)
                            sampler.use(location=unit)
                            samplers.append(sampler)
                        else:
                            child_textures[name].use(location=unit)
                        member.value = unit
                        unit += 1
                    elif name == 'size':
                        member.value = self.size
                    elif name == 'pass':
                        pass_member = member
                    elif name in node:
                        set_uniform_vector(member, node[name])
                    elif name in kwargs:
                        member.value = kwargs[name]
                    elif name in ('alpha', 'gamma'):
                        member.value = 1
            self.glctx.enable_direct(GL_FRAMEBUFFER_SRGB)
            self.glctx.enable(moderngl.BLEND)
            self.glctx.blend_func = moderngl.ONE, moderngl.ONE_MINUS_SRC_ALPHA
            self.framebuffer.use()
            for pass_number in range(passes):
                if self._last is not None:
                    self.glctx.copy_framebuffer(self._last, self.framebuffer)
                if pass_member is not None:
                    pass_member.value = pass_number
                self.framebuffer.clear(*self.CLEAR_COLOR)
                self._rectangle.render()
            for sampler in samplers:
                sampler.clear()
            self.glctx.disable_direct(GL_FRAMEBUFFER_SRGB)


class Shader(ProgramNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._colorbits = None
        self._framebuffer = None
        self._texture = None

    @property
    def texture(self):
        return self._texture

    @property
    def framebuffer(self):
        return self._framebuffer

    def release(self):
        self._colorbits = None
        self._framebuffer = None
        self._texture = None
        super().release()

    def create(self, engine, node, resized, **kwargs):
        super().create(engine, node, resized, **kwargs)
        colorbits = node.get('colorbits', 1, int, self.glctx.extra['colorbits'])
        if colorbits not in COLOR_FORMATS:
            colorbits = self.glctx.extra['colorbits']
        if self._framebuffer is None or self._texture is None or resized or colorbits != self._colorbits:
            depth = COLOR_FORMATS[colorbits]
            self._last = None
            self._texture = self.glctx.texture((self.width, self.height), 4, dtype=depth.moderngl_dtype)
            self._framebuffer = self.glctx.framebuffer(color_attachments=(self._texture,))
            self._framebuffer.clear()
            self._colorbits = colorbits
            logger.debug("Created {}x{} {}-bit framebuffer for {}", self.width, self.height, self._colorbits, self.name)

    def make_last(self):
        logger.debug("Create {}x{} {}-bit last texture for {}", self.width, self.height, self._colorbits, self.name)
        return self.glctx.texture((self.width, self.height), 4, dtype=COLOR_FORMATS[self._colorbits].moderngl_dtype)


class GLFWLoader:
    def load_opengl_function(self, name):
        address = glfw.get_proc_address(name)
        if address is None:
            if name == 'glClearDepth':
                if not hasattr(self, '_glClearDepth'):
                    glClearDepthf = ctypes.cast(glfw.get_proc_address('glClearDepthf'), GLFUNCTYPE(None, ctypes.c_float))

                    def glClearDepth(depth):
                        glClearDepthf(depth)
                    self._glClearDepth = GLFUNCTYPE(None, ctypes.c_double)(glClearDepth)
                return ctypes.cast(self._glClearDepth, ctypes.c_void_p).value
            elif name in ('glReadBuffer', 'glDrawBuffer'):
                if not hasattr(self, '_glIgnore'):
                    def glIgnore(i):
                        pass
                    self._glIgnore = GLFUNCTYPE(None, ctypes.c_double)(glIgnore)
                return ctypes.cast(self._glIgnore, ctypes.c_void_p).value
            else:
                logger.trace("Unresolved GL function: {}", name)
                return 0
        return address

    def release(self):
        pass


class Window(ProgramNode):
    CLEAR_COLOR = (0, 0, 0, 1)
    Windows = []

    def __init__(self, screen=0, fullscreen=False, vsync=False, offscreen=False, **kwargs):
        super().__init__(None)
        self.engine = None
        self.window = None
        self.default_screen = screen
        self.default_fullscreen = fullscreen
        self.default_vsync = vsync
        self._deferred_fullscreen = False
        self._visible = not offscreen
        self._screen = None
        self._fullscreen = None
        self._resizable = None
        self._beat = None

    def release(self):
        if self.glctx is not None:
            self.glctx.finish()
            self.glctx.extra.clear()
            if count := self.glctx.gc():
                logger.trace("Window release collected {} OpenGL objects", count)
            self.glctx.release()
            self.glctx = None
        if self.window is not None:
            glfw.destroy_window(self.window)
            Window.Windows.remove(self)
            if not Window.Windows:
                glfw.terminate()
            logger.debug("{} closed", self.name)
            self.window = None
        self.engine = None
        super().release()

    @property
    def texture(self):
        return None

    @property
    def framebuffer(self):
        return self.glctx.screen

    @property
    def size(self):
        return self.glctx.screen.viewport[2:]

    def create(self, engine, node, resized, opengl_es, **kwargs):
        super().create(engine, node, resized)
        new_window = False
        screen = node.get('screen', 1, int, self.default_screen)
        fullscreen = node.get('fullscreen', 1, bool, self.default_fullscreen) if self._visible else False
        resizable = node.get('resizable', 1, bool, True) if self._visible else False
        if self.window is None:
            self.engine = engine
            title = node.get('title', 1, str, "Flitter")
            if not Window.Windows:
                ok = glfw.init()
                if not ok:
                    raise RuntimeError("Unable to initialize GLFW")
            glfw.window_hint(glfw.CONTEXT_CREATION_API, glfw.EGL_CONTEXT_API if opengl_es else glfw.NATIVE_CONTEXT_API)
            glfw.window_hint(glfw.CLIENT_API, glfw.OPENGL_ES_API if opengl_es else glfw.OPENGL_API)
            glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
            glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 0 if opengl_es else 3)
            glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_ANY_PROFILE if opengl_es else glfw.OPENGL_CORE_PROFILE)
            glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, glfw.TRUE)
            glfw.window_hint(glfw.SRGB_CAPABLE, glfw.TRUE)
            glfw.window_hint(glfw.SAMPLES, 0)
            glfw.window_hint(glfw.VISIBLE, glfw.TRUE if self._visible else glfw.FALSE)
            if self._visible:
                glfw.window_hint(glfw.DOUBLEBUFFER, glfw.TRUE)
                glfw.window_hint(glfw.AUTO_ICONIFY, glfw.FALSE)
                glfw.window_hint(glfw.CENTER_CURSOR, glfw.FALSE)
                glfw.window_hint(glfw.SCALE_TO_MONITOR, glfw.TRUE)
            self.window = glfw.create_window(self.width, self.height, title, None, Window.Windows[0].window if Window.Windows else None)
            Window.Windows.append(self)
            new_window = True
        if self._visible and resizable != self._resizable:
            glfw.set_window_attrib(self.window, glfw.RESIZABLE, glfw.TRUE if resizable else glfw.FALSE)
            self._resizable = resizable
        if self._visible and (resized or screen != self._screen or fullscreen != self._fullscreen):
            monitors = glfw.get_monitors()
            monitor = monitors[screen] if screen < len(monitors) else monitors[0]
            mx, my, mw, mh = glfw.get_monitor_workarea(monitor)
            width, height = self.width, self.height
            if not fullscreen:
                glfw.set_window_aspect_ratio(self.window, self.width, self.height)
            while width > mw * 0.95 or height > mh * 0.95:
                width = width * 2 // 3
                height = height * 2 // 3
            if self._fullscreen and not fullscreen or screen != self._screen:
                glfw.set_window_monitor(self.window, None, 0, 0, width, height, glfw.DONT_CARE)
            if new_window or (self._fullscreen and not fullscreen) or screen != self._screen:
                glfw.set_window_pos(self.window, mx + (mw - width) // 2, my + (mh - height) // 2)
            if fullscreen and (not self._fullscreen or screen != self._screen):
                mode = glfw.get_video_mode(monitor)
                glfw.set_window_monitor(self.window, monitor, 0, 0, mode.size.width, mode.size.height, mode.refresh_rate)
            if not fullscreen:
                glfw.set_window_size(self.window, width, height)
            self._screen = screen
            self._fullscreen = fullscreen
        glfw.make_context_current(self.window)
        self._keys = {}
        self._pointer_state = None
        if new_window:
            moderngl.init_context(GLFWLoader())
            self.glctx = moderngl.get_context()
            self.glctx.gc_mode = 'context_gc'
            self.glctx.extra = {}
            if self._visible:
                logger.debug("{} opened on screen {}", self.name, screen)
            else:
                logger.debug("{} opened off-screen", self.name)
            logger.debug("OpenGL info: {GL_RENDERER} {GL_VERSION}", **self.glctx.info)
            logger.trace("{!r}", self.glctx.info)
            if self._visible:
                glfw.set_key_callback(self.window, self.key_callback)
                glfw.set_cursor_pos_callback(self.window, self.pointer_movement_callback)
                glfw.set_mouse_button_callback(self.window, self.pointer_button_callback)
        if self._visible:
            self.recalculate_viewport(new_window)
        self.glctx.extra['linear'] = node.get('linear', 1, bool, DEFAULT_LINEAR)
        colorbits = node.get('colorbits', 1, int, DEFAULT_COLORBITS)
        if colorbits not in COLOR_FORMATS:
            colorbits = DEFAULT_COLORBITS
        self.glctx.extra['colorbits'] = colorbits
        self.glctx.extra['size'] = self.width, self.height
        if opengl_es:
            self.glctx.extra['HEADER'] = "#version 300 es\nprecision highp float;\n"
        else:
            self.glctx.extra['HEADER'] = "#version 330\n"

    def key_callback(self, window, key, scancode, action, mods):
        if key in self._keys:
            state = self._keys[key]
            if action == glfw.PRESS:
                self.engine.state[state] = 1
                self.engine.state[state.concat(PUSHED)] = 1
                self.engine.state[state.concat(RELEASED)] = 0
                self.engine.state[state.concat(PUSHED_BEAT)] = self._beat
            elif action == glfw.RELEASE:
                self.engine.state[state] = 0
                self.engine.state[state.concat(PUSHED)] = 0
                self.engine.state[state.concat(RELEASED)] = 1
                self.engine.state[state.concat(RELEASED_BEAT)] = self._beat
        elif key == glfw.KEY_LEFT:
            if action == glfw.RELEASE:
                self.engine.previous_page()
        elif key == glfw.KEY_RIGHT:
            if action == glfw.RELEASE:
                self.engine.next_page()

    def pointer_movement_callback(self, window, x, y):
        if self._pointer_state is not None and self.window is not None:
            width, height = glfw.get_window_size(self.window)
            if 0 <= x <= width and 0 <= y <= height:
                self.engine.state[self._pointer_state] = x/width, y/height
            else:
                self.engine.state[self._pointer_state] = None

    def pointer_button_callback(self, window, button, action, mods):
        if self._pointer_state is not None and self.window is not None:
            if action == glfw.PRESS:
                self.engine.state[self._pointer_state.concat(Vector(button))] = 1
            elif action == glfw.RELEASE:
                self.engine.state[self._pointer_state.concat(Vector(button))] = 0

    def handle_node(self, engine, node, **kwargs):
        if node.kind == 'key':
            if 'state' in node:
                key_name = node.get('name', 1, str)
                if key_name:
                    key_constant = 'KEY_' + key_name.upper()
                    if hasattr(glfw, key_constant):
                        key = getattr(glfw, key_constant)
                        state = node['state']
                        self._keys[key] = state
                        if state not in engine.state:
                            if self.window is not None and glfw.get_key(self.window, key) == glfw.PRESS:
                                engine.state[state] = 1
                                engine.state[state.concat(PUSHED)] = 1
                                engine.state[state.concat(RELEASED)] = 0
                            else:
                                engine.state[state] = 0
                                engine.state[state.concat(PUSHED)] = 0
                                engine.state[state.concat(RELEASED)] = 1
            return True
        elif node.kind == 'pointer':
            if 'state' in node:
                self._pointer_state = node['state']
            return True
        return super().handle_node(engine, node, **kwargs)

    def recalculate_viewport(self, force=False):
        aspect_ratio = self.width / self.height
        width, height = glfw.get_framebuffer_size(self.window)
        if width / height > aspect_ratio:
            view_width = int(height * aspect_ratio)
            viewport = ((width - view_width) // 2, 0, view_width, height)
        else:
            view_height = int(width / aspect_ratio)
            viewport = (0, (height - view_height) // 2, width, view_height)
        if force or viewport != self.glctx.screen.viewport:
            self.glctx.screen.viewport = viewport
            width, height = glfw.get_window_size(self.window)
            if viewport != (0, 0, width, height):
                logger.debug("{0} resized to {1}x{2} (viewport {5}x{6} x={3} y={4})", self.name, width, height, *viewport)
            else:
                logger.debug("{} resized to {}x{}", self.name, width, height)
            self._last = None

    def render(self, node, window_gamma=1, beat=None, **kwargs):
        gamma = node.get('gamma', 1, float, window_gamma)
        super().render(node, gamma=gamma, **kwargs)
        if self._visible:
            vsync = node.get('vsync', 1, bool, self.default_vsync)
            glfw.swap_interval(1 if vsync else 0)
            glfw.swap_buffers(self.window)
        self._beat = beat
        glfw.poll_events()
        if count := self.glctx.gc():
            logger.trace("Collected {} OpenGL objects", count)

    def make_last(self):
        width, height = self.window.get_framebuffer_size()
        return self.glctx.texture((width, height), 4)


RENDERER_CLASS = Window


ClassCache = {
    'reference': Reference,
    'shader': Shader,
}



# cython: language_level=3, profile=False

"""
Flitter language functions
"""


import cython

from libc.math cimport floor, round, sin, cos, tan, asin, acos, sqrt, exp, atan2, log, log2, log10
from libc.stdint cimport int64_t, uint64_t
from cpython.object cimport PyObject
from cpython.ref cimport Py_INCREF
from cpython.tuple cimport PyTuple_New, PyTuple_GET_ITEM, PyTuple_SET_ITEM

from ..cache import SharedCache
from ..model cimport Vector, Matrix44, Context, null_, true_, false_


cdef double Pi = 3.141592653589793
cdef double Tau = 6.283185307179586


def context_func(func):
    func.context_func = True
    return func


@context_func
def debug(Context context, Vector value):
    context.logs.add(value.repr())
    return value


@context_func
@cython.boundscheck(False)
def sample(Context context, Vector texture_id, Vector coord, Vector default=null_):
    cdef const double[:, :, :] data
    if coord.length != 2 or coord.numbers == NULL \
            or (scene_node := context.references.get(texture_id.as_string())) is None \
            or not hasattr(scene_node, 'texture_data') or (data := scene_node.texture_data) is None:
        return default
    cdef int64_t height=data.shape[0], width=data.shape[1], x=int(coord.numbers[0] * width), y=int(coord.numbers[1] * height)
    if x < 0 or x >= width or y < 0 or y >= height:
        return default
    cdef const double[:] color = data[y, x]
    cdef Vector result = Vector.__new__(Vector)
    result.allocate_numbers(3)
    result.numbers[0] = color[0]
    result.numbers[1] = color[1]
    result.numbers[2] = color[2]
    return result


cdef class uniform(Vector):
    def __cinit__(self, value=None):
        self._hash = self.hash(True)
        self.deallocate_numbers()
        self.length = 0
        self.objects = None

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self.eq(Vector._coerce(other)).numbers[0] != 0

    cdef Vector eq(self, Vector other):
        if type(other) is self.__class__ and other._hash == self._hash:
            return true_
        return false_

    cdef Vector item(self, int64_t i):
        cdef Vector value = Vector.__new__(Vector)
        value.allocate_numbers(1)
        value.numbers[0] = self._item(i)
        return value

    @cython.cdivision(True)
    cdef double _item(self, uint64_t i) noexcept nogil:
        cdef uint64_t x, y, z
        # Compute a 32bit float PRN using the Squares algorithm [https://arxiv.org/abs/2004.06278]
        x = y = i * self._hash
        z = y + self._hash
        x = x*x + y
        x = (x >> 32) | (x << 32)
        x = x*x + z
        x = (x >> 32) | (x << 32)
        x = x*x + y
        x = (x >> 32) | (x << 32)
        x = x*x + z
        return <double>(x >> 32) / <double>(1<<32)

    cpdef Vector slice(self, Vector index):
        if index.numbers == NULL:
            return null_
        cdef Vector result = Vector.__new__(Vector)
        cdef int64_t i
        cdef uint64_t j
        for i in range(result.allocate_numbers(index.length)):
            j = <uint64_t>(<long long>floor(index.numbers[i]))
            result.numbers[i] = self._item(j)
        return result

    cpdef bint as_bool(self):
        return True

    def __repr__(self):
        return f"{self.__class__.__name__}({self._hash!r})"


cdef class beta(uniform):
    cdef double _item(self, uint64_t i) noexcept nogil:
        i <<= 2
        cdef double u1 = uniform._item(self, i)
        cdef double u2 = uniform._item(self, i + 1)
        cdef double u3 = uniform._item(self, i + 2)
        if u1 <= u2 and u1 <= u3:
            return min(u2, u3)
        if u2 <= u1 and u2 <= u3:
            return min(u1, u3)
        return min(u1, u2)


cdef class normal(uniform):
    @cython.cdivision(True)
    cdef double _item(self, uint64_t i) noexcept nogil:
        # Use the Box-Muller transform to approximate the normal distribution
        # [https://en.wikipedia.org/wiki/Box–Muller_transform]
        cdef double u1, u2
        cdef bint odd = i & 1
        if odd:
            i ^= 1
        if not self.cached or i != self.i:
            u1 = uniform._item(self, i)
            u2 = uniform._item(self, i + 1)
            if u1 < 1 / <double>(1<<32):
                u1, u2 = u2, u1
            self.R = sqrt(-2 * log(u1))
            self.th = Tau * u2
            self.i = i
            self.cached = True
        if odd:
            return self.R * sin(self.th)
        return self.R * cos(self.th)


@context_func
def read_text(Context context, Vector filename):
    cdef str path = filename.as_string()
    if path:
        return Vector._coerce(SharedCache.get_with_root(path, context.path).read_text(encoding='utf8'))
    return null_


@context_func
def read_bytes(Context context, Vector filename):
    cdef str path = filename.as_string()
    if path:
        return Vector._coerce(SharedCache.get_with_root(path, context.path).read_bytes())
    return null_


def ordv(Vector text):
    cdef str string = text.as_string()
    return Vector._coerce([ord(ch) for ch in string])


def chrv(Vector ordinals):
    if ordinals.numbers == NULL:
        return null_
    cdef str text = ""
    cdef int64_t i
    for i in range(ordinals.length):
        text += chr(<int>ordinals.numbers[i])
    cdef Vector result = Vector.__new__(Vector)
    result.objects = (text,)
    result.length = 1
    return result


def split(Vector text):
    if text.length == 0:
        return null_
    return Vector._coerce(text.as_string().rstrip('\n').split('\n'))


@context_func
def read_csv(Context context, Vector filename, Vector row_number):
    cdef str path = str(filename)
    row = row_number.match(1, int)
    if filename and row is not None:
        return SharedCache.get_with_root(path, context.path).read_csv_vector(row)
    return null_


def length(Vector xs not None):
    cdef Vector ys = Vector.__new__(Vector)
    ys.allocate_numbers(1)
    ys.numbers[0] = xs.length
    return ys


def sinv(Vector theta not None):
    if theta.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = theta.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = sin(theta.numbers[i] * Tau)
    return ys


def cosv(Vector theta not None):
    if theta.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = theta.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = cos(theta.numbers[i] * Tau)
    return ys


def tanv(Vector theta not None):
    if theta.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = theta.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = tan(theta.numbers[i] * Tau)
    return ys


def asinv(Vector ys not None):
    if ys.numbers == NULL:
        return null_
    cdef Vector theta = Vector.__new__(Vector)
    cdef int64_t i, n = ys.length
    for i in range(theta.allocate_numbers(n)):
        theta.numbers[i] = asin(ys.numbers[i]) / Tau
    return theta


def acosv(Vector ys not None):
    if ys.numbers == NULL:
        return null_
    cdef Vector theta = Vector.__new__(Vector)
    cdef int64_t i, n = ys.length
    for i in range(theta.allocate_numbers(n)):
        theta.numbers[i] = acos(ys.numbers[i]) / Tau
    return theta


def polar(Vector theta not None):
    if theta.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = theta.length
    ys.allocate_numbers(n*2)
    for i in range(0, n):
        ys.numbers[i*2] = cos(theta.numbers[i] * Tau)
        ys.numbers[i*2+1] = sin(theta.numbers[i] * Tau)
    return ys


def angle(Vector xy not None, Vector ys=None):
    if xy.numbers == NULL or (ys is not None and ys.numbers == NULL):
        return null_
    cdef int64_t i, n = xy.length // 2 if ys is None else max(xy.length, ys.length)
    cdef double x, y
    cdef Vector theta = Vector.__new__(Vector)
    theta.allocate_numbers(n)
    if ys is None:
        for i in range(n):
            x = xy.numbers[i*2]
            y = xy.numbers[i*2+1]
            theta.numbers[i] = atan2(y, x) / Tau
    else:
        for i in range(n):
            x = xy.numbers[i % xy.length]
            y = ys.numbers[i % ys.length]
            theta.numbers[i] = atan2(y, x) / Tau
    return theta


def absv(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    for i in range(ys.allocate_numbers(xs.length)):
        ys.numbers[i] = abs(xs.numbers[i])
    return ys


def expv(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    for i in range(ys.allocate_numbers(xs.length)):
        ys.numbers[i] = exp(xs.numbers[i])
    return ys


def sqrtv(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    for i in range(ys.allocate_numbers(xs.length)):
        ys.numbers[i] = sqrt(xs.numbers[i])
    return ys


def logv(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = xs.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = log(xs.numbers[i])
    return ys


def log2v(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = xs.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = log2(xs.numbers[i])
    return ys


def log10v(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t i, n = xs.length
    for i in range(ys.allocate_numbers(n)):
        ys.numbers[i] = log10(xs.numbers[i])
    return ys


def sine(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    for i in range(ys.allocate_numbers(xs.length)):
        ys.numbers[i] = (1 - cos(Tau * xs.numbers[i])) / 2
    return ys


def bounce(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        ys.numbers[i] = sin(Pi * (x-floor(x)))
    return ys


def impulse(Vector xs not None, Vector cs=None):
    if xs.numbers == NULL or (cs is not None and cs.numbers == NULL):
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t n=cs.length
    cdef double x, c, y, yy
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        x -= floor(x)
        c = cs.numbers[i % n] if cs is not None else 0.25
        if x < c:
            y = 1 - x/c
            yy = y * y
            y = 1 - yy*y
        else:
            y = 1 - (x-c)/(1-c)
            yy = y * y
            y = yy*y
        ys.numbers[i] = y
    return ys


def sharkfin(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        x -= floor(x)
        y = sin(Pi * x) if x < 0.5 else 1 - sin(Pi * (x - 0.5))
        ys.numbers[i] = y
    return ys


def sawtooth(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        ys.numbers[i] = x - floor(x)
    return ys


def triangle(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        x -= floor(x)
        y = 1 - abs(x - 0.5) * 2
        ys.numbers[i] = y
    return ys


def square(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        x -= floor(x)
        y = 0 if x < 0.5 else 1
        ys.numbers[i] = y
    return ys


def linear(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        if x < 0:
            x = 0
        elif x > 1:
            x = 1
        ys.numbers[i] = x
    return ys


def quad(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y, x_
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        if x < 0:
            y = 0
        elif x > 1:
            y = 1
        elif x < 0.5:
            x_ = x * 2
            y = x_*x_ / 2
        else:
            x_ = (1 - x) * 2
            y = 1 - x_*x_ / 2
        ys.numbers[i] = y
    return ys


def cubic(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y, x_
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        if x < 0:
            y = 0
        elif x > 1:
            y = 1
        elif x < 0.5:
            x_ = x * 2
            y = x_*x_*x_ / 2
        else:
            x_ = (1 - x) * 2
            y = 1 - x_*x_*x_ / 2
        ys.numbers[i] = y
    return ys


def snap(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        if x < 0:
            x = 0
        elif x > 1:
            x = 1
        y = sqrt(x * 2) / 2 if x < 0.5 else 1 - sqrt((1 - x) * 2) / 2
        ys.numbers[i] = y
    return ys


cpdef shuffle(uniform source, Vector xs):
    if xs.length == 0:
        return null_
    cdef int64_t i, j, n = xs.length
    cdef PyObject* a
    cdef PyObject* b
    xs = Vector._copy(xs)
    cdef tuple objects=xs.objects
    if objects is None:
        for i in range(n - 1):
            j = <int>floor(source.item(i) * n) + i
            n -= 1
            xs.numbers[i], xs.numbers[j] = xs.numbers[j], xs.numbers[i]
    else:
        for i in range(n - 1):
            j = <int>floor(source.item(i) * n) + i
            n -= 1
            a = PyTuple_GET_ITEM(objects, i)
            b = PyTuple_GET_ITEM(objects, j)
            PyTuple_SET_ITEM(objects, i, <object>b)
            PyTuple_SET_ITEM(objects, j, <object>a)
    return xs


def roundv(Vector xs not None):
    if xs.numbers == NULL:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef double x, y
    for i in range(ys.allocate_numbers(xs.length)):
        x = xs.numbers[i]
        y = round(x)
        ys.numbers[i] = y
    return ys


def ceilv(Vector xs not None):
    return xs.ceil()


def floorv(Vector xs not None):
    return xs.floor()


def fract(Vector xs not None):
    return xs.fract()


def sumv(Vector xs not None, Vector w=true_):
    cdef int64_t i, j, n = xs.length
    if n == 0 or xs.objects is not None or w.length != 1 or w.objects is not None:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t m = <int>(w.numbers[0])
    if m < 1:
        return null_
    ys.allocate_numbers(m)
    for i in range(m):
        ys.numbers[i] = 0
    for i in range(0, n, m):
        for j in range(min(m, n-i)):
            ys.numbers[j] += xs.numbers[i+j]
    return ys


def accumulate(Vector xs not None, Vector zs=true_):
    cdef int64_t i, j, k, n = xs.length
    if n == 0 or xs.objects is not None or zs.length != 1 or zs.objects is not None:
        return null_
    cdef Vector ys = Vector.__new__(Vector)
    cdef int64_t m = <int>(zs.numbers[0])
    if m < 1:
        return null_
    ys.allocate_numbers(n)
    for j in range(min(n, m)):
        ys.numbers[j] = xs.numbers[j]
    for i in range(m, n, m):
        for j in range(min(m, n-i)):
            k = i + j
            ys.numbers[k] = ys.numbers[k-m] + xs.numbers[k]
    return ys


def minv(Vector xs not None, *args):
    cdef Vector ys = null_
    cdef double f
    cdef int64_t i, n = xs.length
    if not args:
        if n:
            if xs.objects is None:
                f = xs.numbers[0]
                for i in range(1, n):
                    if xs.numbers[i] < f:
                        f = xs.numbers[i]
                ys = Vector.__new__(Vector)
                ys.allocate_numbers(1)
                ys.numbers[0] = f
            else:
                y = xs.objects[0]
                for i in range(1, n):
                    x = xs.objects[i]
                    if x < y:
                        y = x
                ys = Vector.__new__(Vector)
                ys.objects = (y,)
                ys.length = 1
    else:
        ys = xs
        for xs in args:
            if xs.compare(ys) == -1:
                ys = xs
    return ys


def minindex(Vector xs not None, *args):
    cdef Vector ys
    cdef double f
    cdef int64_t i, j = 0, n = xs.length
    if not args:
        if n:
            if xs.objects is None:
                f = xs.numbers[0]
                for i in range(1, n):
                    if xs.numbers[i] < f:
                        f = xs.numbers[i]
                        j = i
            else:
                y = xs.objects[0]
                for i in range(1, n):
                    x = xs.objects[i]
                    if x < y:
                        y = x
                        j = i
    else:
        for i, ys in enumerate(args):
            if ys.compare(xs) == -1:
                xs = ys
                j = i + 1
    ys = Vector.__new__(Vector)
    ys.allocate_numbers(1)
    ys.numbers[0] = j
    return ys


def maxv(Vector xs not None, *args):
    cdef Vector ys = null_
    cdef double f
    cdef int64_t i, n = xs.length
    if not args:
        if n:
            if xs.objects is None:
                f = xs.numbers[0]
                for i in range(1, n):
                    if xs.numbers[i] > f:
                        f = xs.numbers[i]
                ys = Vector.__new__(Vector)
                ys.allocate_numbers(1)
                ys.numbers[0] = f
            else:
                y = xs.objects[0]
                for i in range(1, n):
                    x = xs.objects[i]
                    if x > y:
                        y = x
                ys = Vector.__new__(Vector)
                ys.objects = (y,)
                ys.length = 1
    else:
        ys = xs
        for xs in args:
            if xs.compare(ys) == 1:
                ys = xs
    return ys


def maxindex(Vector xs not None, *args):
    cdef Vector ys
    cdef double f
    cdef int64_t i, j = 0, n = xs.length
    if not args:
        if n:
            if xs.objects is None:
                f = xs.numbers[0]
                for i in range(1, n):
                    if xs.numbers[i] > f:
                        f = xs.numbers[i]
                        j = i
            else:
                y = xs.objects[0]
                for i in range(1, n):
                    x = xs.objects[i]
                    if x > y:
                        y = x
                        j = i
    else:
        for i, ys in enumerate(args):
            if ys.compare(xs) == 1:
                xs = ys
                j = i + 1
    ys = Vector.__new__(Vector)
    ys.allocate_numbers(1)
    ys.numbers[0] = j
    return ys


def hypot(*args):
    cdef int64_t i, n = 0
    cdef double x, y = 0
    cdef Vector xs, ys = null_
    if len(args) == 1:
        xs = args[0]
        if xs.numbers == NULL:
            return null_
        for i in range(xs.length):
            x = xs.numbers[i]
            y += x * x
        y = sqrt(y)
        ys = Vector.__new__(Vector)
        ys.allocate_numbers(1)
        ys.numbers[0] = y
    elif len(args) > 1:
        for xs in args:
            if xs.numbers == NULL:
                return null_
            if xs.length > n:
                n = xs.length
        ys = Vector.__new__(Vector)
        ys.allocate_numbers(n)
        for i in range(n):
            y = 0
            for xs in args:
                x = xs.numbers[i % xs.length]
                y += x * x
            ys.numbers[i] = sqrt(y)
    return ys


def normalize(Vector xs not None):
    return xs.normalize()


@cython.cdivision(True)
def mapv(Vector xs not None, Vector ys not None, Vector zs not None):
    if xs.numbers == NULL or ys.numbers == NULL or zs.numbers == NULL:
        return null_
    cdef int64_t i, m=xs.length, n=ys.length, o=zs.length
    cdef Vector ws = Vector.__new__(Vector)
    cdef double x
    for i in range(ws.allocate_numbers(max(m, n, o))):
        x = xs.numbers[i % m]
        ws.numbers[i] = (1-x)*ys.numbers[i % n] + x*zs.numbers[i % o]
    return ws


@cython.cdivision(True)
def clamp(Vector xs not None, Vector ys not None, Vector zs not None):
    if xs.numbers == NULL or ys.numbers == NULL or zs.numbers == NULL:
        return null_
    return xs.clamp(ys, zs)


@cython.cdivision(True)
def zipv(*vectors):
    cdef bint numeric = True
    cdef list vs = []
    cdef Vector v
    cdef int64_t n = 0
    for v in vectors:
        if v.length:
            vs.append(v)
            if v.length > n:
                n = v.length
            numeric = numeric and v.objects is None
    cdef int64_t i, j, p, m = len(vs)
    if m == 0:
        return null_
    if m == 1:
        return vs[0]
    cdef Vector zs = Vector.__new__(Vector)
    cdef double* zp
    cdef object obj
    if numeric:
        zs.allocate_numbers(n * m)
        for j in range(m):
            v = vs[j]
            p = v.length
            zp = zs.numbers + j
            for i in range(n):
                zp[i*m] = v.numbers[i % p]
    else:
        zs.objects = PyTuple_New(n * m)
        for i in range(n):
            for j in range(m):
                v = vs[j]
                if v.objects is None:
                    obj = v.numbers[i % v.length]
                else:
                    obj = <object>PyTuple_GET_ITEM(v.objects, i % v.length)
                Py_INCREF(obj)
                PyTuple_SET_ITEM(zs.objects, i*m + j, obj)
        zs.length = n * m
    return zs


def count(Vector xs not None, Vector ys not None):
    if xs.numbers == NULL or ys.objects is not None:
        return null_
    cdef Vector zs = Vector.__new__(Vector)
    cdef int64_t i, j, k, n=xs.length, m=ys.length
    zs.allocate_numbers(n)
    for i in range(n):
        k = 0
        for j in range(m):
            if ys.numbers[j] == xs.numbers[i]:
                k += 1
        zs.numbers[i] = k
    return zs


cdef double hue_to_rgb(double m1, double m2, double h):
    h = h % 6
    if h < 1:
        return m1 + (m2 - m1) * h
    if h < 3:
        return m2
    if h < 4:
        return m1 + (m2 - m1) * (4 - h)
    return m1


cdef Vector hsl_to_rgb(double h, double s, double l):
    cdef double m2 = l * (s + 1) if l <= 0.5 else l + s - l*s
    cdef double m1 = l * 2 - m2
    h *= 6
    cdef Vector rgb = Vector.__new__(Vector)
    rgb.allocate_numbers(3)
    rgb.numbers[0] = hue_to_rgb(m1, m2, h + 2)
    rgb.numbers[1] = hue_to_rgb(m1, m2, h)
    rgb.numbers[2] = hue_to_rgb(m1, m2, h - 2)
    return rgb


def hsl(Vector color):
    if color.length != 3 or color.objects is not None:
        return null_
    cdef double hue = color.numbers[0], saturation = color.numbers[1], lightness = color.numbers[2]
    saturation = min(max(0, saturation), 1)
    lightness = min(max(0, lightness), 1)
    return hsl_to_rgb(hue, saturation, lightness)


@cython.cdivision(True)
def hsv(Vector color):
    if not color.length or color.length > 3 or color.objects is not None:
        return null_
    cdef double hue = color.numbers[0]
    cdef double saturation = color.numbers[1] if color.length > 1 else 1
    cdef double value = color.numbers[2] if color.length == 3 else 1
    saturation = min(max(0, saturation), 1)
    value = min(max(0, value), 1)
    cdef double lightness = value * (1 - saturation / 2)
    return hsl_to_rgb(hue, 0 if lightness == 0 or lightness == 1 else (value - lightness) / min(lightness, 1 - lightness), lightness)


@cython.cdivision(True)
def colortemp(Vector t):
    if t.numbers == NULL:
        return null_
    cdef int64_t i, n = t.length
    cdef double T, T2, x, x2, y, X, Y, Z
    cdef Vector rgb = Vector.__new__(Vector)
    rgb.allocate_numbers(3*n)
    for i in range(n):
        T = min(max(1667, t.numbers[i]), 25000)
        T2 = T * T
        if T < 4000:
            x = -0.2661239e9/(T*T2) - 0.2343589e6/T2 + 0.8776956e3/T + 0.179910
        else:
            x = -3.0258469e9/(T*T2) + 2.1070379e6/T2 + 0.2226347e3/T + 0.240390
        x2 = x * x
        if T < 2222:
            y = -1.1063814*x*x2 - 1.34811020*x2 + 2.18555832*x - 0.20219683
        elif T < 4000:
            y = -0.9549476*x*x2 - 1.37418593*x2 + 2.09137015*x - 0.16748867
        else:
            y = +3.0817580*x*x2 - 5.87338670*x2 + 3.75112997*x - 0.37001483
        Y = (max(0, t.numbers[i]) / 6503.5)**4
        X = Y * x / y
        Z = Y * (1 - x - y) / y
        rgb.numbers[3*i] = max(0, 3.2406255*X - 1.537208*Y - 0.4986286*Z)
        rgb.numbers[3*i+1] = max(0, -0.9689307*X + 1.8757561*Y + 0.0415175*Z)
        rgb.numbers[3*i+2] = max(0, 0.0557101*X - 0.2040211*Y + 1.0569959*Z)
    return rgb


def point_towards(Vector direction, Vector up):
    cdef Matrix44 matrix = Matrix44._look(Vector([0, 0, 0]), direction, up)
    if matrix is not None:
        return matrix.inverse()
    return null_


STATIC_FUNCTIONS = {
    'abs': Vector(absv),
    'accumulate': Vector(accumulate),
    'acos': Vector(acosv),
    'angle': Vector(angle),
    'asin': Vector(asinv),
    'beta': Vector(beta),
    'bounce': Vector(bounce),
    'ceil': Vector(ceilv),
    'chr': Vector(chrv),
    'clamp': Vector(clamp),
    'colortemp': Vector(colortemp),
    'cos': Vector(cosv),
    'count': Vector(count),
    'cubic': Vector(cubic),
    'exp': Vector(expv),
    'floor': Vector(floorv),
    'fract': Vector(fract),
    'hsl': Vector(hsl),
    'hsv': Vector(hsv),
    'hypot': Vector(hypot),
    'impulse': Vector(impulse),
    'len': Vector(length),
    'linear': Vector(linear),
    'log': Vector(logv),
    'log10': Vector(log10v),
    'log2': Vector(log2v),
    'map': Vector(mapv),
    'max': Vector(maxv),
    'maxindex': Vector(maxindex),
    'min': Vector(minv),
    'minindex': Vector(minindex),
    'normal': Vector(normal),
    'normalize': Vector(normalize),
    'ord': Vector(ordv),
    'point_towards': Vector(point_towards),
    'polar': Vector(polar),
    'quad': Vector(quad),
    'round': Vector(roundv),
    'sawtooth': Vector(sawtooth),
    'sharkfin': Vector(sharkfin),
    'shuffle': Vector(shuffle),
    'sin': Vector(sinv),
    'sine': Vector(sine),
    'snap': Vector(snap),
    'split': Vector(split),
    'sqrt': Vector(sqrtv),
    'square': Vector(square),
    'sum': Vector(sumv),
    'tan': Vector(tanv),
    'triangle': Vector(triangle),
    'uniform': Vector(uniform),
    'zip': Vector(zipv),
}

DYNAMIC_FUNCTIONS = {
    'debug': Vector(debug),
    'sample': Vector(sample),
    'csv': Vector(read_csv),
    'read': Vector(read_text),
    'read_bytes': Vector(read_bytes),
}



"""
Flitter language simplifier unit tests
"""

import unittest

from flitter.model import Vector, Node, StateDict, null, true, false
from flitter.language import functions
from flitter.language.tree import (Literal, Name, Sequence,
                                   Positive, Negative, Ceil, Floor, Fract, Power,
                                   Add, Subtract, Multiply, Divide, FloorDivide, Modulo,
                                   EqualTo, NotEqualTo, LessThan, GreaterThan, LessThanOrEqualTo, GreaterThanOrEqualTo,
                                   Not, And, Or, Xor, Range, Slice, Lookup,
                                   Tag, Attributes, Append,
                                   Let, Call, For, IfElse,
                                   Import, Function, Top, Export,
                                   Binding, PolyBinding, IfCondition)


class SimplifierTestCase(unittest.TestCase):
    def assertSimplifiesTo(self, x, y, state=None, dynamic=None, static=None, with_errors=None, with_names=None):
        x, context = x.simplify(state=state, dynamic=dynamic, static=static, return_context=True)
        if not isinstance(x, Let):
            z = x.simplify(state=state, dynamic=dynamic, static=static)
            self.assertEqual(repr(z), repr(x), msg="Simplification not complete")
        self.assertEqual(repr(x), repr(y))
        self.assertEqual(context.errors, set() if with_errors is None else with_errors)
        if with_names:
            for name, value in with_names.items():
                if isinstance(value, (Name, Function)):
                    self.assertEqual(repr(context.names.pop(name)), repr(value), msg=f"{name} differs from expected")
                else:
                    self.assertEqual(context.names.pop(name), value, msg=f"{name} differs from expected")
        if static is not None:
            for name in static:
                if with_names is None or name not in with_names:
                    self.assertEqual(context.names.pop(name), static[name], msg=f"{name} differs from original static value")
        if dynamic is not None:
            for name in dynamic:
                if with_names is None or name not in with_names:
                    self.assertEqual(context.names.pop(name), None, msg=f"{name} should be dynamic")
        self.assertEqual(len(context.names), 0, msg=f"Unexpected names: {context.names!r}")


class TestLiteral(SimplifierTestCase):
    def test_unchanged(self):
        """Literals are unaffected by simplification."""
        self.assertSimplifiesTo(Literal([1, 2, 3]), Literal([1, 2, 3]))
        self.assertSimplifiesTo(Literal('foo'), Literal('foo'))


class TestName(SimplifierTestCase):
    def test_undefined(self):
        """Undefined names are replaced with literal nulls"""
        self.assertSimplifiesTo(Name('x'), Literal(null), with_errors={"Unbound name 'x'"})

    def test_dynamic(self):
        """Dynamic names are unchanged"""
        self.assertSimplifiesTo(Name('x'), Name('x'), dynamic={'x'})

    def test_static(self):
        """Static Vectors simplify to a Literal"""
        self.assertSimplifiesTo(Name('x'), Literal(5), static={'x': 5})

    def test_rename(self):
        """Static Names simplify to the result of simplifying that name (renaming hack)"""
        self.assertSimplifiesTo(Name('x'), Name('y'), static={'x': Name('y')}, dynamic={'y'})

    def test_function_name(self):
        """Static Functions are left alone (inlining hack)"""
        self.assertSimplifiesTo(Name('f'), Name('f'), static={'f': Function('f', (), Literal(null))})

    def test_static_builtin(self):
        """Static built-ins resolve to their literal value"""
        self.assertSimplifiesTo(Name('null'), Literal(null))
        self.assertSimplifiesTo(Name('sqrt'), Literal(functions.sqrtv))

    def test_dynamic_builtin(self):
        """Dynamic built-ins are left alone"""
        self.assertSimplifiesTo(Name('debug'), Name('debug'))


class TestSequence(SimplifierTestCase):
    def test_empty(self):
        """Empty sequences simplify to literal null"""
        self.assertSimplifiesTo(Sequence(()), Literal(null))

    def test_single(self):
        """Single-item sequences simplify to the single expression"""
        self.assertSimplifiesTo(Sequence((Name('x'),)), Name('x'), dynamic={'x'})

    def test_sequence_packing(self):
        """Sequences within a Sequence are packed together"""
        self.assertSimplifiesTo(Sequence((Name('x'), Sequence((Name('y'), Sequence((Name('y'), Name('y'))))), Sequence((Name('z'),)))),
                                Sequence((Name('x'), Name('y'), Name('y'), Name('y'), Name('z'))),
                                dynamic={'x', 'y', 'z'})

    def test_literal_composition(self):
        """Sequential literal vectors are composed"""
        self.assertSimplifiesTo(Sequence((Name('x'), Literal([1, 2, 3]), Literal([4, 5]), Name('y'))),
                                Sequence((Name('x'), Literal([1, 2, 3, 4, 5]), Name('y'))), dynamic={'x', 'y'})

    def test_recursive(self):
        """Each item in a sequence is simplified"""
        self.assertSimplifiesTo(Sequence((Name('x'), Name('y'))), Literal([1, 2, 3, 4, 5]), static={'x': [1, 2, 3], 'y': [4, 5]})


class TestPositive(SimplifierTestCase):
    def test_numeric_literal(self):
        """Numeric literals are left alone"""
        self.assertSimplifiesTo(Positive(Literal(5)), Literal(5))

    def test_non_numeric_literal(self):
        """Non-numeric literals become nulls"""
        self.assertSimplifiesTo(Positive(Literal('foo')), Literal(null))

    def test_double_positive(self):
        """Double-positives become positive"""
        self.assertSimplifiesTo(Positive(Positive(Name('x'))), Positive(Name('x')), dynamic={'x'})

    def test_positive_negative(self):
        """Positive of a negative becomes the negative"""
        self.assertSimplifiesTo(Positive(Negative(Name('x'))), Negative(Name('x')), dynamic={'x'})

    def test_positive_binary_maths(self):
        """Positive of a binary mathematical operation becomes that operation"""
        self.assertSimplifiesTo(Positive(Add(Name('x'), Name('y'))), Add(Name('x'), Name('y')), dynamic={'x', 'y'})


class TestNegative(SimplifierTestCase):
    def test_numeric_literal(self):
        """Numeric literal gets negated"""
        self.assertSimplifiesTo(Negative(Literal(5)), Literal(-5))

    def test_non_numeric_literal(self):
        """Non-numeric literal becomes null"""
        self.assertSimplifiesTo(Negative(Literal('foo')), Literal(null))

    def test_double_negative(self):
        """Double-negative becomes positive"""
        self.assertSimplifiesTo(Negative(Negative(Name('x'))), Positive(Name('x')), dynamic={'x'})

    def test_multiplication(self):
        """Half-literal multiplication has negative pushed into literal"""
        self.assertSimplifiesTo(Negative(Multiply(Literal(5), Name('x'))), Multiply(Literal(-5), Name('x')), dynamic={'x'})
        # And the other way round:
        self.assertSimplifiesTo(Negative(Multiply(Name('x'), Literal(5))), Multiply(Name('x'), Literal(-5)), dynamic={'x'})

    def test_division(self):
        """Half-literal division has negative pushed into literal"""
        self.assertSimplifiesTo(Negative(Divide(Literal(5), Name('x'))), Divide(Literal(-5), Name('x')), dynamic={'x'})
        # The other way round the division is turned into a multiplication by the inverse of the literal:
        self.assertSimplifiesTo(Negative(Divide(Name('x'), Literal(5))), Multiply(Literal(-0.2), Name('x')), dynamic={'x'})

    def test_addition(self):
        """Half-literal addition becomes a subtraction"""
        # Either way round - this is because of the rule for adding a negative
        self.assertSimplifiesTo(Negative(Add(Literal(5), Name('x'))), Subtract(Literal(-5), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Negative(Add(Name('x'), Literal(5))), Subtract(Literal(-5), Name('x')), dynamic={'x'})

    def test_subtraction(self):
        """Half-literal subtraction becomes an addition"""
        self.assertSimplifiesTo(Negative(Subtract(Literal(5), Name('x'))), Add(Literal(-5), Name('x')), dynamic={'x'})
        # However, rule for adding a negative results in a subtraction the other way round
        self.assertSimplifiesTo(Negative(Subtract(Name('x'), Literal(5))), Subtract(Literal(5), Name('x')), dynamic={'x'})


class TestCeil(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic ceiling is left alone"""
        self.assertSimplifiesTo(Ceil(Name('x')), Ceil(Name('x')), dynamic={'x'})

    def test_literal(self):
        """Literal ceiling is evaluated to a literal"""
        self.assertSimplifiesTo(Ceil(Literal(4.3)), Literal(5))


class TestFloor(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic floor is left alone"""
        self.assertSimplifiesTo(Floor(Name('x')), Floor(Name('x')), dynamic={'x'})

    def test_literal(self):
        """Literal floor is evaluated to a literal"""
        self.assertSimplifiesTo(Floor(Literal(4.3)), Literal(4))


class TestFract(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic floor is left alone"""
        self.assertSimplifiesTo(Fract(Name('x')), Fract(Name('x')), dynamic={'x'})

    def test_literal(self):
        """Literal fract is evaluated to a literal"""
        self.assertSimplifiesTo(Fract(Literal(4.3)), Literal(0.3))


class TestAdd(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Add(Name('x'), Name('y')), Add(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Add(Name('x'), Name('y')), Add(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Add(Name('x'), Name('y')), Add(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Add(Literal(5), Literal(10)), Literal(15))

    def test_zero(self):
        """Adding literal zero becomes Positive"""
        self.assertSimplifiesTo(Add(Literal(0), Name('x')), Positive(Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Add(Name('x'), Literal(0)), Positive(Name('x')), dynamic={'x'})

    def test_negative(self):
        """Adding a Negative becomes a Subtract"""
        self.assertSimplifiesTo(Add(Name('x'), Negative(Name('y'))), Subtract(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(Add(Negative(Name('x')), Name('y')), Subtract(Name('y'), Name('x')), dynamic={'x', 'y'})


class TestSubtract(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Subtract(Name('x'), Name('y')), Subtract(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Subtract(Name('x'), Name('y')), Subtract(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Subtract(Name('x'), Name('y')), Subtract(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Subtract(Literal(5), Literal(10)), Literal(-5))

    def test_subtract_zero(self):
        """Subtracting literal zero becomes Positive"""
        self.assertSimplifiesTo(Subtract(Name('x'), Literal(0)), Positive(Name('x')), dynamic={'x'})

    def test_subtract_from_zero(self):
        """Subtracting from literal zero becomes Negative"""
        self.assertSimplifiesTo(Subtract(Literal(0), Name('x')), Negative(Name('x')), dynamic={'x'})

    def test_negative(self):
        """Subtracting a Negative becomes an Add"""
        self.assertSimplifiesTo(Subtract(Name('x'), Negative(Name('y'))), Add(Name('x'), Name('y')), dynamic={'x', 'y'})


class TestMultiply(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Multiply(Name('x'), Name('y')), Multiply(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Multiply(Name('x'), Name('y')), Multiply(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Multiply(Name('x'), Name('y')), Multiply(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Multiply(Literal(5), Literal(10)), Literal(50))

    def test_multiply_one(self):
        """Multiplying by literal 1 becomes Positive"""
        self.assertSimplifiesTo(Multiply(Name('x'), Literal(1)), Positive(Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(1), Name('x')), Positive(Name('x')), dynamic={'x'})

    def test_multiply_minus_one(self):
        """Multiplying by literal -1 becomes Negative"""
        self.assertSimplifiesTo(Multiply(Name('x'), Literal(-1)), Negative(Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(-1), Name('x')), Negative(Name('x')), dynamic={'x'})

    def test_add_propogation(self):
        """Multiplying a half-literal Add by a literal propogates constant"""
        self.assertSimplifiesTo(Multiply(Add(Name('x'), Literal(5)), Literal(10)), Add(Multiply(Literal(10), Name('x')), Literal(50)), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Add(Name('x'), Literal(5))), Add(Multiply(Literal(10), Name('x')), Literal(50)), dynamic={'x'})

    def test_subtract_propogation(self):
        """Multiplying a half-literal Subtract by a literal propogates constant"""
        self.assertSimplifiesTo(Multiply(Subtract(Literal(5), Name('x')), Literal(10)), Subtract(Literal(50), Multiply(Literal(10), Name('x'))), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Subtract(Literal(5), Name('x'))), Subtract(Literal(50), Multiply(Literal(10), Name('x'))), dynamic={'x'})

    def test_multiply_propogation(self):
        """Multiplying a half-literal Multiply by a literal propogates constant"""
        self.assertSimplifiesTo(Multiply(Multiply(Literal(5), Name('x')), Literal(10)), Multiply(Literal(50), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Multiply(Literal(5), Name('x'))), Multiply(Literal(50), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Multiply(Name('x'), Literal(5)), Literal(10)), Multiply(Literal(50), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Multiply(Name('x'), Literal(5))), Multiply(Literal(50), Name('x')), dynamic={'x'})

    def test_divide_propogation(self):
        """Multiplying a half-literal Divide by a literal propogates constant"""
        self.assertSimplifiesTo(Multiply(Divide(Literal(5), Name('x')), Literal(10)), Divide(Literal(50), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Divide(Literal(5), Name('x'))), Divide(Literal(50), Name('x')), dynamic={'x'})
        # When the Divide denominator is the literal, it is propogated into the Multiply literal
        self.assertSimplifiesTo(Multiply(Divide(Name('x'), Literal(5)), Literal(10)), Multiply(Literal(2), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Divide(Name('x'), Literal(5))), Multiply(Literal(2), Name('x')), dynamic={'x'})

    def test_negative_fold(self):
        """Multiplying a Negative by a literal folds negation into literal"""
        self.assertSimplifiesTo(Multiply(Negative(Name('x')), Literal(10)), Multiply(Literal(-10), Name('x')), dynamic={'x'})
        self.assertSimplifiesTo(Multiply(Literal(10), Negative(Name('x'))), Multiply(Literal(-10), Name('x')), dynamic={'x'})


class TestDivide(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Divide(Name('x'), Name('y')), Divide(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Divide(Name('x'), Name('y')), Divide(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Divide(Name('x'), Name('y')), Divide(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Divide(Literal(5), Literal(10)), Literal(0.5))

    def test_divide_by_one(self):
        """Dividing by literal 1 becomes Positive"""
        self.assertSimplifiesTo(Divide(Name('x'), Literal(1)), Positive(Name('x')), dynamic={'x'})

    def test_divide_by_literal(self):
        """Dividing by literal becomes Multiply of inverse"""
        self.assertSimplifiesTo(Divide(Name('x'), Literal(10)), Multiply(Literal(0.1), Name('x')), dynamic={'x'})


class TestFloorDivide(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(FloorDivide(Name('x'), Name('y')), FloorDivide(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(FloorDivide(Name('x'), Name('y')), FloorDivide(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(FloorDivide(Name('x'), Name('y')), FloorDivide(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(FloorDivide(Literal(5), Literal(10)), Literal(0))

    def test_divide_by_one(self):
        """Dividing by literal 1 becomes Floor"""
        self.assertSimplifiesTo(FloorDivide(Name('x'), Literal(1)), Floor(Name('x')), dynamic={'x'})


class TestModulo(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Modulo(Name('x'), Name('y')), Modulo(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Modulo(Name('x'), Name('y')), Modulo(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Modulo(Name('x'), Name('y')), Modulo(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Modulo(Literal(5), Literal(10)), Literal(5))

    def test_modulo_one(self):
        """Modulo literal 1 becomes Fract"""
        self.assertSimplifiesTo(Modulo(Name('x'), Literal(1)), Fract(Name('x')), dynamic={'x'})


class TestPower(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Power(Name('x'), Name('y')), Power(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Power(Name('x'), Name('y')), Power(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(Power(Name('x'), Name('y')), Power(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(Power(Literal(5), Literal(2)), Literal(25))

    def test_raise_to_power_of_one(self):
        """Power to literal 1 becomes Positive"""
        self.assertSimplifiesTo(Power(Name('x'), Literal(1)), Positive(Name('x')), dynamic={'x'})


class TestEqualTo(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(EqualTo(Name('x'), Name('y')), EqualTo(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(EqualTo(Literal(5), Name('y')), EqualTo(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(EqualTo(Name('x'), Literal(5)), EqualTo(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(EqualTo(Name('x'), Name('y')), EqualTo(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(EqualTo(Name('x'), Name('y')), EqualTo(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(EqualTo(Literal(5), Literal(5)), Literal(true))
        self.assertSimplifiesTo(EqualTo(Literal(5), Literal(4)), Literal(false))
        self.assertSimplifiesTo(EqualTo(Literal(4), Literal(5)), Literal(false))


class TestNotEqualTo(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(NotEqualTo(Name('x'), Name('y')), NotEqualTo(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(NotEqualTo(Literal(5), Name('y')), NotEqualTo(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(NotEqualTo(Name('x'), Literal(5)), NotEqualTo(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(NotEqualTo(Name('x'), Name('y')), NotEqualTo(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(NotEqualTo(Name('x'), Name('y')), NotEqualTo(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(NotEqualTo(Literal(5), Literal(5)), Literal(false))
        self.assertSimplifiesTo(NotEqualTo(Literal(5), Literal(4)), Literal(true))
        self.assertSimplifiesTo(NotEqualTo(Literal(4), Literal(5)), Literal(true))


class TestLessThan(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(LessThan(Name('x'), Name('y')), LessThan(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(LessThan(Literal(5), Name('y')), LessThan(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(LessThan(Name('x'), Literal(5)), LessThan(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(LessThan(Name('x'), Name('y')), LessThan(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(LessThan(Name('x'), Name('y')), LessThan(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(LessThan(Literal(5), Literal(5)), Literal(false))
        self.assertSimplifiesTo(LessThan(Literal(5), Literal(4)), Literal(false))
        self.assertSimplifiesTo(LessThan(Literal(4), Literal(5)), Literal(true))


class TestGreaterThan(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(GreaterThan(Name('x'), Name('y')), GreaterThan(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(GreaterThan(Literal(5), Name('y')), GreaterThan(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(GreaterThan(Name('x'), Literal(5)), GreaterThan(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(GreaterThan(Name('x'), Name('y')), GreaterThan(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(GreaterThan(Name('x'), Name('y')), GreaterThan(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(GreaterThan(Literal(5), Literal(5)), Literal(false))
        self.assertSimplifiesTo(GreaterThan(Literal(5), Literal(4)), Literal(true))
        self.assertSimplifiesTo(GreaterThan(Literal(4), Literal(5)), Literal(false))


class TestLessThanOrEqualTo(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(LessThanOrEqualTo(Name('x'), Name('y')), LessThanOrEqualTo(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(LessThanOrEqualTo(Literal(5), Name('y')), LessThanOrEqualTo(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(LessThanOrEqualTo(Name('x'), Literal(5)), LessThanOrEqualTo(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(LessThanOrEqualTo(Name('x'), Name('y')), LessThanOrEqualTo(Name('x'), Name('z')), dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(LessThanOrEqualTo(Name('x'), Name('y')), LessThanOrEqualTo(Name('z'), Name('y')), dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(LessThanOrEqualTo(Literal(5), Literal(5)), Literal(true))
        self.assertSimplifiesTo(LessThanOrEqualTo(Literal(5), Literal(4)), Literal(false))
        self.assertSimplifiesTo(LessThanOrEqualTo(Literal(4), Literal(5)), Literal(true))


class TestGreaterThanOrEqualTo(SimplifierTestCase):
    def test_dynamic(self):
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Name('x'), Name('y')), GreaterThanOrEqualTo(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Literal(5), Name('y')), GreaterThanOrEqualTo(Literal(5), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Name('x'), Literal(5)), GreaterThanOrEqualTo(Name('x'), Literal(5)), dynamic={'x'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Name('x'), Name('y')), GreaterThanOrEqualTo(Name('x'), Name('z')),
                                dynamic={'x', 'z'}, static={'y': Name('z')})
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Name('x'), Name('y')), GreaterThanOrEqualTo(Name('z'), Name('y')),
                                dynamic={'y', 'z'}, static={'x': Name('z')})

    def test_literal(self):
        """Literal left and right is evaluated"""
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Literal(5), Literal(5)), Literal(true))
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Literal(5), Literal(4)), Literal(true))
        self.assertSimplifiesTo(GreaterThanOrEqualTo(Literal(4), Literal(5)), Literal(false))


class TestNot(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Not(Name('x')), Not(Name('x')), dynamic={'x'})

    def test_recursive(self):
        """Expression is simplified"""
        self.assertSimplifiesTo(Not(Name('x')), Not(Name('y')), dynamic={'y'}, static={'x': Name('y')})

    def test_literal(self):
        """Literal is evaluated"""
        self.assertSimplifiesTo(Not(Literal(false)), Literal(true))
        self.assertSimplifiesTo(Not(Literal(true)), Literal(false))


class TestAnd(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(And(Name('x'), Name('y')), And(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(And(Name('x'), Name('y')), And(Name('z'), Name('y')), static={'x': Name('z')}, dynamic={'y', 'z'})
        self.assertSimplifiesTo(And(Name('x'), Name('y')), And(Name('x'), Name('z')), static={'y': Name('z')}, dynamic={'x', 'z'})

    def test_literal(self):
        """Literal is evaluated"""
        self.assertSimplifiesTo(And(Literal(true), Literal(true)), Literal(true))
        self.assertSimplifiesTo(And(Literal(true), Literal(false)), Literal(false))
        self.assertSimplifiesTo(And(Literal(false), Literal(true)), Literal(false))
        self.assertSimplifiesTo(And(Literal(false), Literal(false)), Literal(false))

    def test_true_left(self):
        """True left shortcuts to right"""
        self.assertSimplifiesTo(And(Literal(true), Name('y')), Name('y'), dynamic={'y'})

    def test_true_right(self):
        """True right ignored"""
        self.assertSimplifiesTo(And(Name('x'), Literal(true)), And(Name('x'), Literal(true)), dynamic={'x'})

    def test_false_left(self):
        """False left shortcuts to left"""
        self.assertSimplifiesTo(And(Literal(false), Name('y')), Literal(false), dynamic={'y'})

    def test_false_right(self):
        """False right ignored"""
        self.assertSimplifiesTo(And(Name('x'), Literal(false)), And(Name('x'), Literal(false)), dynamic={'x'})


class TestOr(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Or(Name('x'), Name('y')), Or(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Or(Name('x'), Name('y')), Or(Name('z'), Name('y')), static={'x': Name('z')}, dynamic={'y', 'z'})
        self.assertSimplifiesTo(Or(Name('x'), Name('y')), Or(Name('x'), Name('z')), static={'y': Name('z')}, dynamic={'x', 'z'})

    def test_literal(self):
        """Literal is evaluated"""
        self.assertSimplifiesTo(Or(Literal(true), Literal(true)), Literal(true))
        self.assertSimplifiesTo(Or(Literal(true), Literal(false)), Literal(true))
        self.assertSimplifiesTo(Or(Literal(false), Literal(true)), Literal(true))
        self.assertSimplifiesTo(Or(Literal(false), Literal(false)), Literal(false))

    def test_true_left(self):
        """True left shortcuts to left"""
        self.assertSimplifiesTo(Or(Literal(true), Name('y')), Literal(true), dynamic={'y'})

    def test_true_right(self):
        """True right ignored"""
        self.assertSimplifiesTo(Or(Name('x'), Literal(true)), Or(Name('x'), Literal(true)), dynamic={'x'})

    def test_false_left(self):
        """False left shortcuts to right"""
        self.assertSimplifiesTo(Or(Literal(false), Name('y')), Name('y'), dynamic={'y'})

    def test_false_right(self):
        """False right ignored"""
        self.assertSimplifiesTo(Or(Name('x'), Literal(false)), Or(Name('x'), Literal(false)), dynamic={'x'})


class TestXor(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Xor(Name('x'), Name('y')), Xor(Name('x'), Name('y')), dynamic={'x', 'y'})

    def test_recursive(self):
        """Left and right are simplified"""
        self.assertSimplifiesTo(Xor(Name('x'), Name('y')), Xor(Name('z'), Name('y')), static={'x': Name('z')}, dynamic={'y', 'z'})
        self.assertSimplifiesTo(Xor(Name('x'), Name('y')), Xor(Name('x'), Name('z')), static={'y': Name('z')}, dynamic={'x', 'z'})

    def test_literal(self):
        """Literal is evaluated"""
        self.assertSimplifiesTo(Xor(Literal(true), Literal(true)), Literal(false))
        self.assertSimplifiesTo(Xor(Literal(true), Literal(false)), Literal(true))
        self.assertSimplifiesTo(Xor(Literal(false), Literal(true)), Literal(true))
        self.assertSimplifiesTo(Xor(Literal(false), Literal(false)), Literal(false))

    def test_true_left(self):
        """True left ignored"""
        self.assertSimplifiesTo(Xor(Literal(true), Name('y')), Xor(Literal(true), Name('y')), dynamic={'y'})

    def test_true_right(self):
        """True right ignored"""
        self.assertSimplifiesTo(Xor(Name('x'), Literal(true)), Xor(Name('x'), Literal(true)), dynamic={'x'})

    def test_false_left(self):
        """False left shortcuts to right"""
        self.assertSimplifiesTo(Xor(Literal(false), Name('y')), Name('y'), dynamic={'y'})

    def test_false_right(self):
        """False right shorcuts to left"""
        self.assertSimplifiesTo(Xor(Name('x'), Literal(false)), Name('x'), dynamic={'x'})


class TestRange(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Range(Name('w'), Name('x'), Name('y')), Range(Name('w'), Name('x'), Name('y')), dynamic={'w', 'x', 'y'})

    def test_recursive(self):
        """Start, stop and step are simplified"""
        self.assertSimplifiesTo(Range(Name('w'), Name('x'), Name('y')), Range(Name('z'), Name('x'), Name('y')),
                                dynamic={'x', 'y', 'z'}, static={'w': Name('z')})
        self.assertSimplifiesTo(Range(Name('w'), Name('x'), Name('y')), Range(Name('w'), Name('z'), Name('y')),
                                dynamic={'w', 'y', 'z'}, static={'x': Name('z')})
        self.assertSimplifiesTo(Range(Name('w'), Name('x'), Name('y')), Range(Name('w'), Name('x'), Name('z')),
                                dynamic={'w', 'x', 'z'}, static={'y': Name('z')})

    def test_literal(self):
        """Literal Range is evaluated"""
        self.assertSimplifiesTo(Range(Literal(0), Literal(10), Literal(2)), Literal([0, 2, 4, 6, 8]))


class TestSlice(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Slice(Name('x'), Name('y')), Slice(Name('x'), Name('y')), dynamic={'x', 'y'})
        self.assertSimplifiesTo(Slice(Literal(range(1, 6)), Name('y')), Slice(Literal(range(1, 6)), Name('y')), dynamic={'y'})
        self.assertSimplifiesTo(Slice(Name('x'), Literal(3)), Slice(Name('x'), Literal(3)), dynamic={'x'})

    def test_recursive(self):
        """Expression and index are simplified"""
        self.assertSimplifiesTo(Slice(Name('x'), Name('y')), Slice(Name('z'), Name('y')), static={'x': Name('z')}, dynamic={'y', 'z'})
        self.assertSimplifiesTo(Slice(Name('x'), Name('y')), Slice(Name('x'), Name('z')), static={'y': Name('z')}, dynamic={'x', 'z'})

    def test_literal(self):
        """Literal expr and key is evaluated"""
        self.assertSimplifiesTo(Slice(Literal(range(1, 6)), Literal(3)), Literal(4))


class TestLookup(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic is left alone"""
        self.assertSimplifiesTo(Lookup(Name('x')), Lookup(Name('x')), dynamic={'x'})

    def test_recursive(self):
        """Key is simplified"""
        self.assertSimplifiesTo(Lookup(Name('x')), Lookup(Name('y')), dynamic={'y'}, static={'x': Name('y')})

    def test_literal_key_no_state(self):
        """Literal key with no supplied state is left alone"""
        self.assertSimplifiesTo(Lookup(Literal(Vector.symbol('foo'))), Lookup(Literal(Vector.symbol('foo'))))

    def test_literal_key_not_in_state(self):
        """Literal key not in supplied state is left alone"""
        self.assertSimplifiesTo(Lookup(Literal(Vector.symbol('foo'))), Lookup(Literal(Vector.symbol('foo'))), state=StateDict())

    def test_literal_key_in_state(self):
        """Literal key in supplied state is simplified to literal value"""
        self.assertSimplifiesTo(Lookup(Literal(Vector.symbol('foo'))), Literal(5), state=StateDict({Vector.symbol('foo'): 5}))


class TestTag(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic node expression is left alone"""
        self.assertSimplifiesTo(Tag(Name('node'), 'tag'), Tag(Name('node'), 'tag'), dynamic={'node'})

    def test_literal_node(self):
        """Literal node is tagged"""
        self.assertSimplifiesTo(Tag(Literal(Node('node')), 'tag'), Literal(Node('node', {'tag'})))

    def test_literal_nodes(self):
        """Literal nodes are tagged"""
        self.assertSimplifiesTo(Tag(Literal([Node('node1'), Node('node2', {'tag1'})]), 'tag2'),
                                Literal([Node('node1', {'tag2'}), Node('node2', {'tag1', 'tag2'})]))

    def test_literal_null(self):
        """Tag of null becomes null"""
        self.assertSimplifiesTo(Tag(Literal(null), 'tag'), Literal(null))

    def test_literal_numeric(self):
        """Tag of numeric is discarded"""
        self.assertSimplifiesTo(Tag(Literal(5), 'tag'), Literal(5))

    def test_literal_mixed(self):
        """Tag of mixed literal has all nodes updated"""
        self.assertSimplifiesTo(Tag(Literal([5, "hello", Node('node')]), 'tag'), Literal([5, "hello", Node('node', {'tag'})]))


class TestAttributes(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic node or value expressions are left alone"""
        self.assertSimplifiesTo(Attributes(Name('x'), (Binding('y', Literal(5)),)), Attributes(Name('x'), (Binding('y', Literal(5)),)), dynamic={'x'})
        self.assertSimplifiesTo(Attributes(Literal(Node('node')), (Binding('y', Name('y')),)),
                                Attributes(Literal(Node('node')), (Binding('y', Name('y')),)), dynamic={'y'})

    def test_literal_node(self):
        """Literal node has attributes updated"""
        self.assertSimplifiesTo(Attributes(Literal(Node('node')), (Binding('y', Literal(5)),)),
                                Literal(Node('node', attributes={'y': Vector(5)})))

    def test_literal_nodes(self):
        """Literal nodes have attributes updated"""
        self.assertSimplifiesTo(Attributes(Literal([Node('node1'), Node('node2', attributes={'x': Vector(1)})]), (Binding('y', Literal(5)),)),
                                Literal([Node('node1', attributes={'y': Vector(5)}), Node('node2', attributes={'x': Vector(1), 'y': Vector(5)})]))

    def test_partial_application(self):
        """Literal bindings are applied until first dynamic binding"""
        self.assertSimplifiesTo(Attributes(Literal(Node('node')), (Binding('y', Literal(5)), Binding('z', Name('z')))),
                                Attributes(Literal(Node('node', attributes={'y': Vector(5)})), (Binding('z', Name('z')),)), dynamic={'z'})

    def test_combining(self):
        """Nested attributes are combined"""
        self.assertSimplifiesTo(Attributes(Attributes(Name('node'), (Binding('x', Name('x')),)), (Binding('y', Name('y')), Binding('z', Name('z')))),
                                Attributes(Name('node'), (Binding('x', Name('x')), Binding('y', Name('y')), Binding('z', Name('z')))),
                                dynamic={'node', 'x', 'y', 'z'})

    def test_literal_null(self):
        """Attribute set of null becomes null whether dynamic or not"""
        self.assertSimplifiesTo(Attributes(Literal(null), (Binding('x', Name('x')),)), Literal(null), dynamic={'x'})

    def test_literal_numeric(self):
        """Attribute set of numeric is discarded whether dynamic or not"""
        self.assertSimplifiesTo(Attributes(Literal(5), (Binding('x', Name('x')),)), Literal(5), dynamic={'x'})

    def test_literal_mixed(self):
        """Literal attribute set of mixed literal has all nodes updated"""
        self.assertSimplifiesTo(Attributes(Literal([5, "hello", Node('node')]), (Binding('x', Literal(5)),)),
                                Literal([5, "hello", Node('node', attributes={'x': Vector(5)})]))


class TestAppend(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic node expressions are left alone"""
        self.assertSimplifiesTo(Append(Name('x'), Literal(Node('y'))), Append(Name('x'), Literal(Node('y'))), dynamic={'x'})
        self.assertSimplifiesTo(Append(Literal(Node('x')), Name('y')), Append(Literal(Node('x')), Name('y')), dynamic={'y'})

    def test_literal_node(self):
        """Literal node has sub-node appended"""
        self.assertSimplifiesTo(Append(Literal(Node('x')), Literal(Node('y'))),
                                Literal(Node('x', children=(Node('y'),))))

    def test_literal_nodes(self):
        """Literal nodes have attributes updated"""
        self.assertSimplifiesTo(Append(Literal([Node('x1'), Node('x2', children=(Node('y1'),))]), Literal(Node('y'))),
                                Literal([Node('x1', children=(Node('y'),)), Node('x2', children=(Node('y1'), Node('y')))]))

    def test_push_through_attributes_to_literal(self):
        """Literal appends are pushed through intermediate attribute operation to a literal root"""
        self.assertSimplifiesTo(Append(Attributes(Literal(Node('node1')), (Binding('x', Name('x')),)), Literal(Node('node2'))),
                                Attributes(Literal(Node('node1', children=(Node('node2'),))), (Binding('x', Name('x')),)),
                                dynamic={'x'})

    def test_pull_literal_from_sequence(self):
        """A literal at the start of an appended sequence is pulled out and appended to a literal node"""
        self.assertSimplifiesTo(Append(Literal(Node('node1')), Sequence((Literal(Node('node2')), Name('x'), Name('y')))),
                                Append(Literal(Node('node1', children=(Node('node2'),))), Sequence((Name('x'), Name('y')))),
                                dynamic={'x', 'y'})

    def test_append_to_literal_null(self):
        """Append to null is discarded"""
        self.assertSimplifiesTo(Append(Literal(null), Name('x')), Literal(null), dynamic={'x'})

    def test_append_of_literal_null(self):
        """Append of null is discarded"""
        self.assertSimplifiesTo(Append(Name('x'), Literal(null)), Name('x'), dynamic={'x'})

    def test_append_to_literal_numeric(self):
        """Append to numeric is discarded"""
        self.assertSimplifiesTo(Append(Literal(5), Name('x')), Literal(5), dynamic={'x'})

    def test_append_of_literal_numeric(self):
        """Append of numeric is discarded"""
        self.assertSimplifiesTo(Append(Name('x'), Literal(5)), Name('x'), dynamic={'x'})

    def test_literal_mixed(self):
        """Mixed literal append to mixed literal has all nodes updated"""
        self.assertSimplifiesTo(Append(Literal([5, "hello", Node('node')]), Literal([Node('node2'), 7])),
                                Literal([5, "hello", Node('node', children=(Node('node2'),))]))


class TestLet(SimplifierTestCase):
    def test_dynamic(self):
        """Binding to a dynamic expression is left alone"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Add(Name('y'), Literal(5))),), Add(Name('x'), Name('y'))),
                                Let((PolyBinding(('x',), Add(Name('y'), Literal(5))),), Add(Name('x'), Name('y'))),
                                dynamic={'y'})

    def test_let_in_literal(self):
        """A let over a literal is always just the literal"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Add(Name('y'), Literal(5))),), Literal(10)), Literal(10), dynamic={'y'})

    def test_literal_binding(self):
        """Simple binding of a name to a literal"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Literal(5)),), Add(Name('x'), Name('y'))),
                                Add(Literal(5), Name('y')),
                                dynamic={'y'})

    def test_literal_multi_binding(self):
        """Binding of a name sequence to a literal"""
        self.assertSimplifiesTo(Let((PolyBinding(('x', 'y'), Literal([5, 10])),), Add(Name('x'), Name('y'))), Literal(15))

    def test_literal_short_multi_binding(self):
        """Binding of a name sequence to a short literal wraps"""
        self.assertSimplifiesTo(Let((PolyBinding(('x', 'y', 'z'), Literal([5, 10])),), Name('z')), Literal(5))

    def test_function_inlining(self):
        """Binding a name to a Function expression places Function into names for inlining"""
        self.assertSimplifiesTo(Let((PolyBinding(('f',), Function('f', (Binding('x', Literal(null)),), Name('x'))),), Call(Name('f'), (Literal(5),))),
                                Literal(5))

    def test_non_literal_function_parameters(self):
        """A function with non-literal default parameter values is not inlined"""
        self.assertSimplifiesTo(Let((PolyBinding(('f',), Function('f', (Binding('x', Name('y')),), Name('x'))),), Call(Name('f'), (Literal(5),))),
                                Let((PolyBinding(('f',), Function('f', (Binding('x', Name('y')),), Name('x'), ())),), Call(Name('f'), (Literal(5),))),
                                dynamic={'y'})

    def test_function_with_captures(self):
        """A function with non-literal default parameter values is not inlined"""
        self.assertSimplifiesTo(Let((PolyBinding(('f',), Function('f', (Binding('x', Literal(null)),), Name('y'))),), Call(Name('f'), (Literal(5),))),
                                Let((PolyBinding(('f',), Function('f', (Binding('x', Literal(null)),), Name('y'), ('y',))),), Call(Name('f'), (Literal(5),))),
                                dynamic={'y'})

    def test_rename(self):
        """Simple rename of a local"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Name('y')),), Add(Name('x'), Name('y'))),
                                Add(Name('y'), Name('y')),
                                dynamic={'y'})

    def test_expr_shadowed_rename(self):
        """Rename of a local that is shadowed by a later binding to an expression"""
        self.assertSimplifiesTo(Let((PolyBinding(('y',), Add(Name('y'), Literal(5))),), Add(Name('x'), Name('y'))),
                                Let((PolyBinding(('x',), Name('y')), PolyBinding(('y',), Add(Name('y'), Literal(5)))), Add(Name('x'), Name('y'))),
                                static={'x': Name('y')}, dynamic={'y'})

    def test_expr_shadowed_rename_subexpr(self):
        """Rename of a local that is shadowed by a binding to an expression in a sub-expression"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Name('y')),),
                                    Add(Literal(5), Let((PolyBinding(('y',), Add(Name('y'), Literal(5))),), Add(Name('x'), Name('y'))))),
                                Add(Literal(5), Let((PolyBinding(('x',), Name('y')),
                                                     PolyBinding(('y',), Add(Name('y'), Literal(5))),), Add(Name('x'), Name('y')))),
                                dynamic={'y'})

    def test_literal_shadowed_rename(self):
        """Rename of a local that is shadowed by a later binding to a literal"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Name('y')), PolyBinding(('y',), Literal(5))), Add(Name('x'), Name('y'))),
                                Add(Name('y'), Literal(5)),
                                dynamic={'y'})

    def test_rename_shadowed_rename(self):
        """Rename of a local that is shadowed by a later binding to a rename"""
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Name('y')), PolyBinding(('y',), Name('z'))), Add(Name('x'), Name('y'))),
                                Add(Name('y'), Name('z')),
                                dynamic={'y', 'z'})


class TestCall(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic calls left alone"""
        self.assertSimplifiesTo(Call(Name('x'), (Literal(5),)), Call(Name('x'), (Literal(5),)), dynamic={'x'})
        self.assertSimplifiesTo(Call(Literal(functions.sqrtv), (Name('y'),)), Call(Literal(functions.sqrtv), (Name('y'),)), dynamic={'y'})
        self.assertSimplifiesTo(Call(Literal(functions.sqrtv), (), (Binding('xs', Name('y')),)),
                                Call(Literal(functions.sqrtv), (), (Binding('xs', Name('y')),)), dynamic={'y'})

    def test_static(self):
        """Static calls to built-in functions are replaced with their return value"""
        self.assertSimplifiesTo(Call(Literal(functions.sqrtv), (Literal(25),), ()), Literal(5))
        self.assertSimplifiesTo(Call(Literal(functions.sqrtv), (), (Binding('xs', Literal(25)),)), Literal(5))

    def test_simple_inlining(self):
        """Calls to names that resolve to Function objects are inlined as let expressions"""
        func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)), captures=(), inlineable=True)
        self.assertSimplifiesTo(Call(Name('func'), (Add(Literal(1), Name('y')),), ()),
                                Let((PolyBinding(('x',), Add(Literal(1), Name('y'))),), Add(Name('x'), Literal(5))),
                                static={'func': func}, dynamic={'y'})

    def test_inlineable_recursive_non_literal(self):
        """Calls to inlineable, recursive functions are *not* inlined if arguments are not all literal"""
        func = Function(
            'func',
            (Binding('x', Literal(null)),),
            IfElse((IfCondition(GreaterThan(Name('x'), Literal(0)), Add(Name('x'), Call(Name('func'), (Subtract(Name('x'), Literal(1)),)))),), Literal(0)),
            captures=(), inlineable=True, recursive=True
        )
        self.assertSimplifiesTo(Call(Name('func'), (Name('y'),)), Call(Name('func'), (Name('y'),)), static={'func': func}, dynamic={'y'})

    def test_inlineable_recursive_literal(self):
        """Calls to inlineable, recursive functions *are* inlined if arguments are all literal"""
        func = Function(
            'func',
            (Binding('x', Literal(null)),),
            IfElse((IfCondition(GreaterThan(Name('x'), Literal(0)), Add(Name('x'), Call(Name('func'), (Subtract(Name('x'), Literal(1)),)))),), Literal(0)),
            captures=(), inlineable=True, recursive=True
        )
        self.assertSimplifiesTo(Call(Name('func'), (Literal(5),)), Literal(15), static={'func': func})


class TestFor(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic loop source left alone"""
        self.assertSimplifiesTo(For(('x',), Name('y'), Name('x')), For(('x',), Name('y'), Name('x')), dynamic={'y'})

    def test_recursive(self):
        """Source and body simplified"""
        self.assertSimplifiesTo(For(('x',), Name('y'), Add(Name('x'), Name('z'))),
                                For(('x',), Name('z'), Add(Name('x'), Name('z'))),
                                static={'y': Name('z')}, dynamic={'z'})
        self.assertSimplifiesTo(For(('x',), Name('y'), Add(Name('x'), Name('z'))),
                                For(('x',), Name('y'), Add(Name('x'), Name('w'))),
                                static={'z': Name('w')}, dynamic={'y', 'w'})

    def test_single_name_unroll(self):
        """Simple iteration of a single name over a literal vector"""
        self.assertSimplifiesTo(For(('x',), Literal([1, 2]), Add(Name('x'), Name('z'))),
                                Sequence((Add(Literal(1), Name('z')), Add(Literal(2), Name('z')))),
                                dynamic={'z'})

    def test_multiple_name_unroll(self):
        """Iteration of multiple names over a literal vector"""
        self.assertSimplifiesTo(For(('x', 'y'), Literal([1, 2, 3]), Call(Name('f'), (Name('x'), Name('y')), ())),
                                Sequence((Call(Name('f'), (Literal(1), Literal(2))), Call(Name('f'), (Literal(3), Literal(null))))),
                                dynamic={'f'})


class TestIfElse(SimplifierTestCase):
    def test_dynamic(self):
        """Dynamic condition left alone"""
        self.assertSimplifiesTo(IfElse((IfCondition(Name('x'), Literal(5)),), None), IfElse((IfCondition(Name('x'), Literal(5)),), None), dynamic={'x'})

    def test_recursive(self):
        """Sub expressions are all simplified"""
        self.assertSimplifiesTo(IfElse((IfCondition(Name('x'), Name('y')),), Name('z')),
                                IfElse((IfCondition(Name('w'), Name('y')),), Name('z')),
                                static={'x': Name('w')}, dynamic={'w', 'y', 'z'})
        self.assertSimplifiesTo(IfElse((IfCondition(Name('x'), Name('y')),), Name('z')),
                                IfElse((IfCondition(Name('x'), Name('w')),), Name('z')),
                                static={'y': Name('w')}, dynamic={'x', 'w', 'z'})
        self.assertSimplifiesTo(IfElse((IfCondition(Name('x'), Name('y')),), Name('z')),
                                IfElse((IfCondition(Name('x'), Name('y')),), Name('w')),
                                static={'z': Name('w')}, dynamic={'x', 'y', 'w'})

    def test_true_condition(self):
        """A true condition simplifies to the then expression"""
        self.assertSimplifiesTo(IfElse((IfCondition(Literal(true), Name('y')),), Name('z')), Name('y'), dynamic={'y', 'z'})

    def test_false_condition_else(self):
        """A false condition simplifies to the else expression"""
        self.assertSimplifiesTo(IfElse((IfCondition(Literal(false), Name('y')),), Name('z')), Name('z'), dynamic={'y', 'z'})

    def test_false_condition_no_else(self):
        """A false condition without an else simplifies to null"""
        self.assertSimplifiesTo(IfElse((IfCondition(Literal(false), Name('y')),), None), Literal(null), dynamic={'y'})

    def test_false_condition_1_of_2(self):
        """A false 1st condition is removed when more than one condition"""
        self.assertSimplifiesTo(IfElse((IfCondition(Literal(false), Name('x')), IfCondition(Name('w'), Name('y'))), Name('z')),
                                IfElse((IfCondition(Name('w'), Name('y')),), Name('z')), dynamic={'w', 'x', 'y', 'z'})

    def test_true_condition_2_of_2(self):
        """A true 2nd condition simplifies to an if with the 2nd then as the else"""
        self.assertSimplifiesTo(IfElse((IfCondition(Name('w'), Name('x')), IfCondition(Literal(true), Name('y'))), Name('z')),
                                IfElse((IfCondition(Name('w'), Name('x')),), Name('y')), dynamic={'w', 'x', 'y', 'z'})

    def test_false_condition_2_of_2(self):
        """A false 2nd condition is removed"""
        self.assertSimplifiesTo(IfElse((IfCondition(Name('w'), Name('x')), IfCondition(Literal(false), Name('y')),), Name('z')),
                                IfElse((IfCondition(Name('w'), Name('x')),), Name('z')), dynamic={'w', 'x', 'y', 'z'})

    def test_true_condition_2_of_3(self):
        """A true 2nd condition simplifies to an if with the 2nd then as the else; 3rd condition removed"""
        self.assertSimplifiesTo(IfElse((IfCondition(Name('w'), Name('x')),
                                        IfCondition(Literal(true), Name('y')),
                                        IfCondition(Name('a'), Name('b'))), Name('z')),
                                IfElse((IfCondition(Name('w'), Name('x')),), Name('y')), dynamic={'w', 'x', 'y', 'z', 'a', 'b'})


class TestImport(SimplifierTestCase):
    def test_recursive(self):
        """Imports are left alone except for the sub-expression being simplified"""
        self.assertSimplifiesTo(Import(('x', 'y'), Name('m'), Name('z')), Import(('x', 'y'), Literal('module.fl'), Literal(5)),
                                static={'m': 'module.fl', 'z': 5})


class TestFunction(SimplifierTestCase):
    def test_simple_inlineable(self):
        """A simple, unsimplifiable function with no external references will have empty captures and will be inlineable"""
        start_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)))
        simpl_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)), captures=(), inlineable=True)
        self.assertSimplifiesTo(start_func, simpl_func)

    def test_recursive(self):
        """Parameter defaults and the body of the function are simplified (again, this is an inlineable function)"""
        start_func = Function('func', (Binding('x', Name('y')),), Add(Name('x'), Name('z')))
        simpl_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Literal(5)), captures=(), inlineable=True)
        self.assertSimplifiesTo(start_func, simpl_func, static={'y': null, 'z': 5})

    def test_non_literal_default(self):
        """A function that has a dynamic parameter default won't be inlineable"""
        start_func = Function('func', (Binding('x', Name('y')),), Add(Name('x'), Literal(5)))
        simpl_func = Function('func', (Binding('x', Name('y')),), Add(Name('x'), Literal(5)), captures=(), inlineable=False)
        self.assertSimplifiesTo(start_func, simpl_func, dynamic={'y'})

    def test_simple_capture(self):
        """A function that references an external name will have that noted in its captures and won't be inlineable"""
        start_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Name('y')))
        simpl_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Name('y')), captures=('y',))
        self.assertSimplifiesTo(start_func, simpl_func)

    def test_simple_recursive(self):
        """A function that references only itself will be marked as recursive and will be inlineable"""
        start_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Call(Name('func'), (Name('x'),))))
        simpl_func = Function('func', (Binding('x', Literal(null)),), Add(Name('x'), Call(Name('func'), (Name('x'),))),
                              captures=(), inlineable=True, recursive=True)
        self.assertSimplifiesTo(start_func, simpl_func)


class TestExport(SimplifierTestCase):
    def test_empty(self):
        self.assertSimplifiesTo(Export(), Export())

    def test_explicit(self):
        self.assertSimplifiesTo(Export(), Export({'x': Vector(5)}), static={'x': 5})

    def test_let_explicit(self):
        self.assertSimplifiesTo(Let((PolyBinding(('x',), Literal(5)),), Export()), Export({'x': Vector(5)}))


class TestTop(SimplifierTestCase):
    def test_recursive(self):
        """Top expression is simplified"""
        self.assertSimplifiesTo(Top((), Name('x')), Top((), Name('y')), static={'x': Name('y')}, dynamic={'y'})



"""
Flitter language integration tests
"""

import math
from pathlib import Path
import unittest

from flitter import configure_logger
from flitter.language.tree import Literal, Let, Sequence, Export, Binding, Top
from flitter.language.parser import parse
from flitter.language.vm import Function
from flitter.model import Vector, Node, Context, StateDict, DummyStateDict, null


configure_logger('WARNING')


class TestPragmas(unittest.TestCase):
    def test_no_pragmas(self):
        top = parse("\n")
        self.assertEqual(top.pragmas, ())
        program = top.compile()
        self.assertEqual(program.pragmas, {})

    def test_pragmas(self):
        top = parse("""
%pragma foo 5
%pragma bar 'five'
%pragma baz :five
%pragma buz !five
""")
        self.assertEqual(repr(top), repr(Top((Binding('foo', Literal(5)),
                                              Binding('bar', Literal('five')),
                                              Binding('baz', Literal(Vector.symbol('five'))),
                                              Binding('buz', Literal(Node('five')))), Sequence((Export(None),)))))
        program = top.compile()
        self.assertEqual(program.pragmas, {'foo': Vector(5), 'bar': Vector('five'), 'baz': Vector.symbol('five'), 'buz': Vector(Node('five'))})


class TestLanguageFeatures(unittest.TestCase):
    """
    Test a series of language features by parsing the given code, evaluating it and comparing the
    output to that given. The code is evaluated in two ways:

    1. Compile the unsimplified AST and run the resulting VM program with any additional bound
       names passed in as initial lvalues
    2. Simplify the AST with any additional bound names passed in as static, then extract the
       literal values out of the resulting AST

    It is assumed that all of the examples can be fully reduced to a literal by the simplifier.
    """

    def assertCodeOutput(self, code, output, **names):
        top = parse(code.strip() + "\n")
        output = output.strip()
        run_context = Context(names={name: Vector(value) for name, value in names.items()}, state=StateDict())
        vm_output = '\n'.join(repr(node) for node in top.compile(initial_lnames=tuple(names)).run(run_context).root.children)
        self.assertEqual(vm_output, output, msg="VM output is incorrect")
        expr = top.simplify(static=names).body
        while not isinstance(expr, Literal):
            if isinstance(expr, Let):
                expr = expr.body
            elif isinstance(expr, Sequence) and len(expr.expressions) == 2 and isinstance(expr.expressions[1], Export):
                expr = expr.expressions[0]
            else:
                break
        self.assertIsInstance(expr, Literal, msg="Unexpected simplification output")
        simplifier_output = '\n'.join(repr(node) for node in expr.value)
        self.assertEqual(simplifier_output, output, msg="Simplifier output is incorrect")

    def test_literal_node(self):
        self.assertCodeOutput(
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """,
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """)

    def test_names(self):
        self.assertCodeOutput(
            """
!foo x=x y=y z=z #bar
            """,
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """, x=5, y=Vector.symbol('baz'), z='Hello world!')

    def test_let(self):
        self.assertCodeOutput(
            """
let x=5 y=:baz z='Hello world!'
!foo #bar x=x y=y z=z
            """,
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """)

    def test_multibind_loop(self):
        self.assertCodeOutput(
            """
for x;y in ..9
    !node x=x y=y
            """,
            """
!node x=0 y=1
!node x=2 y=3
!node x=4 y=5
!node x=6 y=7
!node x=8
            """)

    def test_nested_loops(self):
        self.assertCodeOutput(
            """
for i in ..n
    !group i=i
        for j in ..i
            !item numbers=(k*(j+1) for k in 1..i+1)
            """,
            """
!group i=0
!group i=1
 !item numbers=1
!group i=2
 !item numbers=1;2
 !item numbers=2;4
!group i=3
 !item numbers=1;2;3
 !item numbers=2;4;6
 !item numbers=3;6;9
            """, n=4)

    def test_if(self):
        self.assertCodeOutput(
            """
for i in ..2
    if i
        !one
            """,
            """
!one
            """)

    def test_if_else(self):
        self.assertCodeOutput(
            """
for i in ..4
    if not i
        !zero
    elif i == 2
        !two
    elif i > 2
        !more
    else
        !one
            """,
            """
!zero
!one
!two
!more
            """)

    def test_recursive_inlined_function(self):
        # Also default parameter values and out-of-order named arguments
        self.assertCodeOutput(
            """
func fib(n, x=0, y=1)
    fib(n-1, y=y+x, x=y) if n > 0 else x

!fib x=fib(10)
            """,
            """
!fib x=55
            """)

    def test_builtin_calls(self):
        self.assertCodeOutput(
            """
for i in ..n
    !color rgb=hsv(i/n;1;1)
            """,
            """
!color rgb=1;0;0
!color rgb=1;1;0
!color rgb=0;1;0
!color rgb=0;1;1
!color rgb=0;0;1
!color rgb=1;0;1
            """, n=6)

    def test_template_calls(self):
        self.assertCodeOutput(
            """
func foo(nodes, x=10, y, z='world')
    !foo x=x z='hello';z
        nodes #test y=y

@foo
@foo z='you'
    !bar
@foo y=5
    @foo z='me' x=99
        !baz
            """,
            """
!foo x=10 z='hello';'world'
!foo x=10 z='hello';'you'
 !bar #test
!foo x=10 z='hello';'world'
 !foo #test x=99 z='hello';'me' y=5
  !baz #test
            """)

    def test_nested_functions(self):
        self.assertCodeOutput(
            """
func fib(n)
    func fib'(n, x, y)
        fib'(n-1, y=y+x, x=y) if n > 0 else x
    fib'(n, 0, 1)

!fib x=fib(10)
            """,
            """
!fib x=55
            """)

    def test_anonymous_functions(self):
        """Note that this is only statically reducible because `map` is inlined and so the
           anonymous function is bound to `f` and therefore becomes a name"""
        self.assertCodeOutput(
            """
func map(f, xs)
    for x in xs
        f(x)

!doubled x=map(func(x, y=2) x*y, ..10)
            """,
            """
!doubled x=0;2;4;6;8;10;12;14;16;18
            """)


class TestExampleScripts(unittest.TestCase):
    """
    Run all Flitter scripts in `/examples`, `/docs/diagrams` and `/docs/tutorial_images` under three
    different conditions and test that the generated root nodes and exported values match:

    1. Compile the AST and run it without *any* simplification and with an empty state
    2. Simplify the AST with known names (`beat`, etc.) as dynamic and an empty state
    3. Simplify the AST with known names as static values and a special dummy state that claims to
       contain all keys but returns null for all as an empty state would

    In the third case, all state lookups will be statically evaluated to literal nulls by the
    simplifier. The idea is to exercise the simplifier as much as possible.
    """

    def assertAllAlmostEqual(self, xs, ys, msg=None):
        for x, y in zip(xs, ys):
            self.assertEqual(math.isnan(x), math.isnan(y), msg=msg)
            if not (math.isnan(x) or math.isnan(y)):
                self.assertAlmostEqual(x, y, msg=msg)

    def assertExportsMatch(self, exports1, exports2, msg=None):
        self.assertEqual(exports1.keys(), exports2.keys())
        for name in exports1:
            a = exports1[name]
            b = exports2[name]
            if a.isinstance(Function) and b.isinstance(Function):
                self.assertEqual(a[0].__name__, b[0].__name__, msg=msg)
            elif a.numeric and b.numeric:
                self.assertAllAlmostEqual(a, b, msg=msg)
            else:
                self.assertEqual(a, b, msg=msg)

    def _test_integration_script(self, filepath):
        self.maxDiff = None
        state = StateDict()
        null_state = DummyStateDict()
        names = {'beat': Vector(0), 'tempo': Vector(120), 'quantum': Vector(4), 'fps': Vector(60), 'OUTPUT': null}
        top = parse(filepath.read_text(encoding='utf8'))
        # Compile un-simplified AST and execute program:
        program1 = top.compile(initial_lnames=tuple(names))
        program1.set_path(filepath)
        program1.use_simplifier(False)
        context = program1.run(Context(names=dict(names), state=state))
        root1 = context.root
        exports1 = context.exports
        windows = [node for node in root1.children if node.kind == 'window']
        self.assertEqual(len(windows), 1, msg="Should be a single window node in program output")
        self.assertGreater(len(tuple(windows[0].children)), 0, "Output window node should have children")
        # Simplify AST with dynamic names, compile and execute:
        top2 = top.simplify(dynamic=set(names))
        self.assertEqual(repr(top2.simplify(dynamic=set(names))), repr(top2), msg="Dynamic simplification not complete in one step")
        program2 = top2.compile(initial_lnames=tuple(names))
        program2.set_path(filepath)
        self.assertNotEqual(len(program1), len(program2), msg="Dynamically-simplified program length should be different from original")
        context = program2.run(Context(names=dict(names), state=state))
        root2 = context.root
        exports2 = context.exports
        self.assertEqual(repr(root2), repr(root1), msg="Dynamically-simplified program output doesn't match original")
        self.assertExportsMatch(exports2, exports1, msg="Dynamically-simplified program exports don't match original")
        # Simplify AST with static names and null state, compile and execute:
        top3 = top.simplify(static=names, state=null_state)
        self.assertEqual(repr(top3.simplify(static=names, state=null_state)), repr(top3), msg="Static simplification not complete in one step")
        program3 = top3.compile()
        program3.set_path(filepath)
        self.assertNotEqual(len(program3), len(program1), msg="Statically-simplified program length should be different from original")
        context = program3.run(Context(state=state))
        root3 = context.root
        exports3 = context.exports
        for name in names:
            del exports3[name]
        self.assertEqual(repr(root3), repr(root1), msg="Statically-simplified program output doesn't match original")
        self.assertExportsMatch(exports3, exports1, msg="Statically-simplified program exports don't match original")

    def _test_integration_all_scripts(self, dirpath):
        count = 0
        for filename in dirpath.iterdir():
            filepath = dirpath / filename
            if filepath.suffix == '.fl':
                with self.subTest(filepath=filepath):
                    self._test_integration_script(filepath)
                    count += 1
        self.assertGreater(count, 0, msg="No scripts found")

    def test_integration_examples(self):
        self._test_integration_all_scripts(Path(__file__).parent.parent / 'examples')

    def test_integration_docs_diagrams(self):
        self._test_integration_all_scripts(Path(__file__).parent.parent / 'docs/diagrams')

    def test_integration_docs_tutorial(self):
        self._test_integration_all_scripts(Path(__file__).parent.parent / 'docs/tutorial_images')
